#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  1 21:57:54 2020
@author: Ajit Johnson Nirmal
Using Napari to Visualize images overlayed with phenotypes or any categorical column
"""

#%gui qt
import napari
import pandas as pd
import random
import tifffile as tiff

def image_viewer (image_path, adata, overlay=None,
                    overlay_category=None,markers=None,channel_names='default',
                    x_coordinate='X_centroid',y_coordinate='Y_centroid',point_size=10,
                    point_color=None,subset=None,imageid='imageid',seg_mask=None):
    """
    Parameters
    ----------
    image_path : string
        Location to the image file.
    seg_mask: string (The default is None)
        Location to the segmentation mask file.
    adata : AnnData Object
    overlay : string, optional (The default is None)
        Name of the column with any categorical data such as phenotypes or clusters.
    overlay_category : list, optional (The default is None)
        If only specfic categories within the overlay column is needed, pass their names as a list.
        If None, all categories will be used.
    markers : list, optional (The default is None)
        Markers to be included. If none, all markers will be displayed.
    channel_names : list, optional (The default is `adata.uns['all_markers']`)
        List of channels in the image in the exact order as image.
    x_coordinate : string, optional (The default is 'X_centroid')
        X axis coordinate column name in AnnData object.
    y_coordinate : string, optional (The default is 'Y_centroid')
        Y axis coordinate column name in AnnData object.
    point_size : int, optional (The default is 10)
        point size in the napari plot.
    imageid : string, optional *(The default is `imageid`)*   
        Column name of the column containing the image id. 
    subset : string, optional  *(The default is None)*  
        imageid of a single image to be subsetted for analyis. Only useful when multiple images are being analyzed together.

    Returns
    -------
    None.

    Example
    -------
    image_path = '/Users/aj/Desktop/ptcl_tma/image.tif'
    sm.pl.image_viewer (image_path, adata, overlay='phenotype',overlay_category=None,
                markers=['CD31', "CD3D","DNA11",'CD19','CD45','CD163','FOXP3'],
                point_size=7,point_color='white')

    """
    # Plot only the Image that is requested
    if subset is not None:
        adata = adata[adata.obs[imageid] == subset]

    # Recover the channel names from adata
    if channel_names is 'default':
        channel_names = adata.uns['all_markers']
    else:
        channel_names = channel_names

    # Index of the marker of interest and corresponding names
    if markers is None:
        idx = list(range(len(channel_names)))
        channel_names = channel_names
    else:
        idx = []
        for i in markers:
            idx.append(list(channel_names).index(i))
        channel_names = markers


    # Load the image
    image = tiff.imread(image_path, key = idx)

    # Load the segmentation mask
    if seg_mask is not None:
        seg_m = tiff.imread(seg_mask)

    # Load the viewer
    viewer = napari.view_image(
    image,
    #is_pyramid=False,
    channel_axis=0,
    name = None if channel_names is None else channel_names,
    visible = False)

    # Add the seg mask
    if seg_mask is not None:
        viewer.add_labels(seg_m, name='segmentation mask')

    # Add phenotype layer function
    def add_phenotype_layer (adata, overlay, phenotype_layer,x,y,viewer,point_size,point_color):
        coordinates = adata[adata.obs[overlay] == phenotype_layer]
        coordinates = pd.DataFrame({'y': coordinates.obs[y],'x': coordinates.obs[x]})
        points = coordinates.values.tolist()
        if point_color is None:
            r = lambda: random.randint(0,255) # random color generator
            point_color = '#%02X%02X%02X' % (r(),r(),r()) # random color generator
        viewer.add_points(points, size=point_size,face_color=point_color,visible=False,name=phenotype_layer)

    if overlay is not None:
        # categories under investigation
        if overlay_category is None:
            available_phenotypes = list(adata.obs[overlay].unique())
        else:
            available_phenotypes = overlay_category

        # Run the function on all phenotypes
        for i in available_phenotypes:
            add_phenotype_layer (adata=adata, overlay=overlay,
                                    phenotype_layer=i, x=x_coordinate, y=y_coordinate, viewer=viewer,
                                    point_size=point_size,point_color=point_color)



from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution('scimap').version
except DistributionNotFound:
    __version__ = '(local)'


from . import preprocessing as pp
from . import plotting as pl
from . import tools as tl
from . import helpers as hl


# -*- coding: utf-8 -*-
"""
Created on Fri Mar  6 12:13:22 2020
@author: Ajit Johnson Nirmal
ReScale the data
"""

# Import library
import os
import pandas as pd
import numpy as np
import itertools
from sklearn.preprocessing import MinMaxScaler
from sklearn.mixture import GaussianMixture
from sklearn.linear_model import LinearRegression
import matplotlib.pyplot as plt
import seaborn as sns; sns.set(color_codes=True)
from matplotlib.lines import Line2D


def rescale (adata, gate=None, return_gates=False, imageid='imageid', failed_markers=None, method='all',save_fig=False):
    
    """


    Parameters
    ----------
    adata : AnnData object
    gate : dataframe, optional (The default is None)
        DataFrame with first column as markers and second column as the gate values in log1p scale.
        Note: If a marker is not included, the function will try to automatically identify a gate
        based on gaussian mixture modeling. If a marker is included in the `gate` dataframe but 
        no values are passed, the marker is simply scaled between 0-1 but does not alter the undelying
        distribution. 
    return_gates : boolian, optional (The default is False)
        Internal parameter for checking.
    failed_markers : list, optional (The default is None)
        list of markers that are not expressed at all in any cell. pass in as ['CD20', 'CD3D'].
    method : string, optional (The default is 'all')
        Two avialble option are- 'all' or 'by_image'. In the event that multiple images were loaded in with distinct 'imageid',
        users have the option to scale all data togeather or each image independently. Please be aware of batch effects when
        passing 'all' with multiple images.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    save_fig : boolian, optional (The default is False)
        If True, the gates identified by the GMM method will be saved in a subdirectory
        within your working directory.

    Returns
    -------
    Modified AnnData Object

    Example
    -------
    manual_gate = pd.DataFrame({'marker': ['CD3D', 'KI67'], 'gate': [7, 8]})
    adata = sm.pp.rescale (adata, gate=manual_gate, failed_markers=['CD20', 'CD21'])

    """


    def rescale_independent (adata, gate, return_gates, failed_markers,save_fig):

        print('Scaling Image '+ str(adata.obs[imageid].unique()))

        # Copy of the raw data if it exisits
        if adata.raw is not None:
            adata.X = adata.raw.X

        data = pd.DataFrame(adata.X, columns = adata.var.index, index= adata.obs.index)
        # Merging the manual gates and non-working markers togeather if any
        if gate is not None:
            m_markers = list(gate.iloc[:,0])
            manual_gate_markers = gate
        if failed_markers != None:
            manual_gate_markers = pd.DataFrame(data[failed_markers].quantile(0.9999999))
            manual_gate_markers['markers'] = failed_markers
            # move column to front
            cols = manual_gate_markers.columns.tolist()
            cols.insert(0, cols.pop(cols.index('markers')))
            manual_gate_markers = manual_gate_markers.reindex(columns= cols)
            manual_gate_markers.columns = ['marker', 'gate']
            m_markers = failed_markers
        if gate is not None and failed_markers != None:
            m_markers = list(gate.iloc[:,0]) + list(manual_gate_markers.iloc[:,0])
            gate.columns = ['marker', 'gate']
            manual_gate_markers = pd.concat([gate, manual_gate_markers])
        if gate is None and failed_markers == None:
            m_markers = []

        # Find markers to send to gmm modelling
        if gate is not None or failed_markers is not None:
            gmm_markers = list(np.setdiff1d(data.columns, m_markers))
        else:
            gmm_markers = list(data.columns)

        # If manual gate is not provided scale the data
        if len(gmm_markers) != 0:
            gmm_data = data[gmm_markers]
            # Clip off the 99th percentile
            def clipping (x):
                clip = x.clip(lower =np.percentile(x,1), upper=np.percentile(x,99)).tolist()
                return clip
            # Run the function
            gmm_data = gmm_data.apply(clipping)

            # Scaling the data old
            sum_data = gmm_data.sum(axis=1) # Calculate total count for each cell
            n_count = gmm_data.div(sum_data, axis=0) # Divide genes by total count for every cell
            med = np.median(list(itertools.chain(*gmm_data.values.tolist()))) # Calculate median count of the entire dataset
            n_count_med = n_count*med # Multiply by scaling fator (median count of entire dataset)
            n_log = np.log1p(n_count_med) # Log transform data
            scaler = MinMaxScaler(feature_range=(0, 1))
            s = scaler.fit_transform(n_log)
            normalised_data = pd.DataFrame(s, columns = gmm_data.columns, index= gmm_data.index)


            # Gaussian fit to identify the gate for each marker and scale based on the gate
            # Empty data frame to hold the results
            all_gmm_data = pd.DataFrame()
            def gmm_gating (data, marker, return_gates, save_fig, gmm_data):
                # Print
                print('Finding the optimal gate for ' + str(marker))
                # Identify the marker to fit the model
                m = data[marker].values
                # Perform GMM
                data_gm = m.reshape(-1, 1)
                #gmm = GaussianMixture(n_components=2, means_init=[[0],[1]],covariance_type='tied')
                gmm = GaussianMixture(n_components=2)
                gmm.fit(data_gm)
                gate = np.mean(gmm.means_)

                # Find the closest value to the gate
                absolute_val_array = np.abs(m - gate)
                smallest_difference_index = absolute_val_array.argmin()
                closest_element = m[smallest_difference_index]

                # If Save Figure is true
                if save_fig == True:
                    m_ndata = gmm_data[marker].values
                    # generate a linear regression for prediction
                    x = data[marker].values.reshape(-1, 1)
                    y = gmm_data[marker].values.reshape(-1, 1)
                    reg = LinearRegression().fit(x, y)
                    #reg.score(x, y)
                    #SVR
                    #reg = SVR(kernel='poly').fit(x, y)
                    #regr.score(x, y)

                    # the three lines
                    g1 = m[np.abs(m - gmm.means_[0]).argmin()]
                    mg = m[np.abs(m - np.mean(gmm.means_)).argmin()]
                    g2 = m[np.abs(m - gmm.means_[1]).argmin()]
                    # predicted gates on log scale
                    g1 = np.log1p(reg.predict(np.array([[g1]])))
                    mg = np.log1p(reg.predict(np.array([[mg]])))
                    g2 = np.log1p(reg.predict(np.array([[g2]])))

                    # saving figure folder
                    if not os.path.exists('auto_gating'):
                        os.makedirs('auto_gating')

                    # generate figure
                    plt.ioff()
                    lines = [Line2D([0], [0], color='b', linestyle='--'),
                             Line2D([1], [1], color='red'),
                             Line2D([0], [0], color='b',linestyle='--')]
                    labels = ['Neg Gaussian', 'Gate', 'Pos Gaussian']
                    sns.set_style("white")
                    fig, ax = plt.subplots( nrows=1, ncols=1 )
                    sns.distplot(np.log1p(m_ndata),color="grey")
                    plt.axvline(g1,ls='--')
                    plt.axvline(g2,ls='--')
                    plt.axvline(mg,color="red")
                    plt.legend(lines, labels)
                    plt.title(marker, fontsize=30)
                    fig.savefig('auto_gating/' + str(marker) + '.png')
                    plt.clf()
                    plt.close('all')
                    plt.ion()

                # rescale the data based on the identified gate
                marker_study = pd.DataFrame(m, index= data.index)
                marker_study = marker_study.sort_values(0)

                # Find the index of the gate
                gate_index = marker_study.index[marker_study[0] == closest_element][0]

                # Split into high and low groups
                high = marker_study.loc[gate_index:,:]
                low = marker_study.loc[:gate_index,:]

                # Prepare for scaling the high and low dataframes
                scaler_high = MinMaxScaler(feature_range=(0.5, 1))
                scaler_low = MinMaxScaler(feature_range=(0, 0.5))

                # Scale it
                h = pd.DataFrame(scaler_high.fit_transform(high), index = high.index)
                l = pd.DataFrame(scaler_low.fit_transform(low), index = low.index)

                # Merge the high and low and resort it
                scaled_data = pd.concat([l,h])
                scaled_data = scaled_data.loc[~scaled_data.index.duplicated(keep='first')]
                scaled_data = scaled_data.reindex(data.index)

                #return scaled_data
                if return_gates == True:
                    return gate
                else:
                    return scaled_data

            # Apply the function
            r_gmm_gating = lambda x: gmm_gating(data=normalised_data, marker=x,return_gates=return_gates,save_fig=save_fig,gmm_data=gmm_data) # Create lamda function
            all_gmm_data = list(map(r_gmm_gating, gmm_markers)) # Apply function
            all_gmm_data = pd.concat(all_gmm_data, axis=1, sort=False)
            all_gmm_data.columns = gmm_markers
        else:
            all_gmm_data = pd.DataFrame()

        # Empty data frame to hold the results
        all_manual_data = pd.DataFrame()
        if len(m_markers) != 0:
            m_data = np.log1p(data[m_markers])
            # Clip the data
            def clipping (x):
                clip = x.clip(lower =np.percentile(x,1), upper=np.percentile(x,99)).tolist()
                return clip
            # Run the function
            m_data = m_data.apply(clipping)

            def manual_gating (data,marker,gate):
             
                # Work on processing manual gates
                m = gate[gate.iloc[:,0] == marker].iloc[:,1].values[0] # gate of the marker passed in
                
                if np.isnan(m):
                    # Find the mean value of the marker so that it is scaled right at the middle
                    # in other it retains the original scale
                    m = np.mean(data[marker].values)
                    print('Warning: No manual gate was found for ' + str(marker) + '. Scaling it between 0 and 1')
                else:
                    print('Scaling ' + str(marker))
                    

                # Find the closest value to the gate
                absolute_val_array = np.abs(data[marker].values - float(m))
                smallest_difference_index = absolute_val_array.argmin()
                closest_element = data[marker].values[smallest_difference_index]

                # rescale the data based on the identified gate
                marker_study = data[marker]
                marker_study = marker_study.sort_values(0)

                # Find the index of the gate
                gate_index = marker_study.index[marker_study == closest_element][0]

                # Split into high and low groups
                high = marker_study[gate_index:]
                low = marker_study[:gate_index]

                # Prepare for scaling the high and low dataframes
                scaler_high = MinMaxScaler(feature_range=(0.5, 1))
                scaler_low = MinMaxScaler(feature_range=(0, 0.5))

                # Scale it
                h = pd.DataFrame(scaler_high.fit_transform(high.values.reshape(-1, 1)), index = high.index)
                l = pd.DataFrame(scaler_low.fit_transform(low.values.reshape(-1, 1)), index = low.index)

                # Merge the high and low and resort it
                scaled_data = pd.concat([l,h])
                scaled_data = scaled_data.loc[~scaled_data.index.duplicated(keep='first')]
                scaled_data = scaled_data.reindex(data.index)

                # Return
                return scaled_data

            # Apply the function
            r_manual_gating = lambda x: manual_gating(data=m_data, marker=x, gate=manual_gate_markers) # Create lamda function
            all_manual_data = list(map(r_manual_gating, m_markers)) # Apply function
            all_manual_data = pd.concat(all_manual_data, axis=1, sort=False)
            all_manual_data.columns = m_markers

        else:
            all_manual_data = pd.DataFrame()


        # If both manual and automatic gating was used, combine them into a single result
        if not all_manual_data.empty:
            all_scaled_data = all_manual_data
        if not all_gmm_data.empty:
            all_scaled_data = all_gmm_data
        if not all_manual_data.empty and not all_gmm_data.empty:
            all_scaled_data = all_gmm_data.merge(all_manual_data, how='outer', left_index=True, right_index=True)

        # re index the columns
        all_scaled_data = all_scaled_data.reindex(columns= data.columns)
        return all_scaled_data

    # Apply method of choice
    if method == 'all':
        all_scaled_data = rescale_independent (adata, gate=gate, return_gates=return_gates, failed_markers=failed_markers, save_fig=save_fig)
    if method == 'by_image':
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
        r_rescale_independent = lambda x: rescale_independent(adata=x, gate=gate, return_gates=return_gates, failed_markers=failed_markers,save_fig=save_fig) # Create lamda function
        scaled_data = list(map(r_rescale_independent, adata_list)) # Apply function
        all_scaled_data = pd.concat(scaled_data)

    # Create a copy of the raw data
    if adata.raw is None:
        adata.raw = adata

    # Replace with normalized data
    adata.X = all_scaled_data

    # Return data
    return adata



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 10 09:32:49 2020
@author: Ajit Johnson Nirmal
Tests
"""

import pytest
import sys, os
import anndata as ad

#os.chdir ("/Users/aj/Dropbox (Partners HealthCare)/packages/scimap")

@pytest.fixture
def adata():
    image_path = os. getcwd() + '/scimap/tests/_data/example_data.h5ad'
    adata = ad.read(image_path)
    return adata

# Testing the phenotyping function
def test_phenotype(adata):
    import pandas as pd
    from scimap.tools._phenotype_cells import phenotype_cells

    # Load phenotype and test phenotyping
    phenotype = pd.read_csv(os. getcwd() + '/scimap/tests/_data/phenotype_workflow.csv')
    adata = phenotype_cells (adata, phenotype=phenotype, gate = 0.5, label="phenotype")
    a = adata.obs['phenotype'][0]

    # test
    assert a == 'M2 Macrophages'

# Testing the spatial_count function
def test_spatial_count(adata):
    from scimap.tools._spatial_count import spatial_count
    
    adata = spatial_count (adata,x_coordinate='X_position',y_coordinate='Y_position',
                           phenotype='phenotype',method='radius',radius=30,
                           imageid='ImageId',subset=None,label='spatial_count_radius')
    a = round(adata.uns['spatial_count_radius']['ASMA+ cells'].sum(),0)
    
    # test
    assert a == 330
    
# Testing the spatial_expression function
def test_spatial_expression(adata):
    from scimap.tools._spatial_expression import spatial_expression
    
    adata = spatial_expression (adata, x_coordinate='X_position',y_coordinate='Y_position',
                                method='radius', radius=30, imageid='ImageId', 
                                use_raw=True,subset=None,label='spatial_expression')
    a = round(adata.uns['spatial_expression']['CD3D'].sum(),0)
    
    # test
    assert a == 19920
    
# Testing the spatial_aggregate function
def test_spatial_aggregate(adata):
    from scimap.tools._spatial_aggregate import spatial_aggregate
    
    adata = spatial_aggregate (adata, x_coordinate='X_position',y_coordinate='Y_position',
                           purity = 60, phenotype='phenotype', method='radius', radius=30,
                           imageid='ImageId',subset=None,label='spatial_aggregate')
    a = adata.obs['spatial_aggregate'].value_counts()['M2 Macrophages']
    
    # test
    assert a == 20
    
# Testing cluster function
def test_cluster(adata):
    from scimap.tools._cluster import cluster
    adata = cluster (adata,  method = 'kmeans', k= 5, use_raw = True)
    a = adata.obs['kmeans'].value_counts()[4]
    
    #test
    assert a == 252

# Testing spatial_interaction function
def test_spatial_interaction(adata):
    from scimap.tools._spatial_interaction import spatial_interaction
    adata = spatial_interaction (adata,  method = 'knn', knn= 5, permutation = 10, imageid='ImageId',x_coordinate='X_position',y_coordinate='Y_position')
    a = adata.uns['spatial_interaction']
    
    #test
    assert a is not None
    
# Testing spatial_distance function
def test_spatial_distance(adata):
    from scimap.tools._spatial_distance import spatial_distance
    adata = spatial_distance (adata, imageid='ImageId',x_coordinate='X_position',y_coordinate='Y_position')
    a = adata.uns['spatial_distance']
    
    #test
    assert a is not None





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 12 23:47:52 2020
@author: Ajit Johnson Nirmal
View the gating in Napari viewer
"""

import napari
import pandas as pd
import tifffile as tiff
import numpy as np

def gate_finder (image_path, adata, marker_of_interest, from_gate = 6, to_gate = 8, increment = 0.1,
                 markers=None, channel_names = 'default',
                 x_coordinate='X_centroid',y_coordinate='Y_centroid',
                 point_size=10,imageid='imageid',subset=None,seg_mask=None):
    """


    Parameters
    ----------
    image_path : string
        Location to the image file..
    adata : Ann Data Object
    marker_of_interest : string
        Marker for which gate is to be defined e.g. 'CD45'.
    from_gate : int, optional (The default is 6)
        Start value gate of interest.
    to_gate : int, optional (The default is 8)
        End value of the gate of interest.
    increment : float, optional (The default is 0.1)
        Increments between the start and end values.
    markers : string, optional (The default is None)
        Additional markers to be included in the plot for evaluation.
    channel_names : list, optional (The default is `adata.uns['all_markers']`)
        List of channels in the image in the exact order as image.
    x_coordinate : string, optional (The default is 'X_centroid')
        X axis coordinate column name in AnnData object.
    y_coordinate : string, optional (The default is 'Y_centroid')
        Y axis coordinate column name in AnnData object.
    point_size : int, optional (The default is 10)
        point size in the napari plot.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    subset : string, optional
        imageid of a single image to be subsetted for analyis. The default is None.
    seg_mask : string, optional (The default is None)
        Location to the segmentation mask file.

    Example
    -------
    image_path = '/Users/aj/Desktop/ptcl_tma/image.tif'
    sm.pl.gate_finder (image_path, adata, marker_of_interest='CD45',
                 from_gate = 6, to_gate = 8, increment = 0.1,
                 markers=['DNA10'], channel_names = 'default',
                 x_coordinate='X_position',y_coordinate='Y_position',point_size=10,
                 subset= '77', seg_mask=None)

    """



    # If no raw data is available make a copy
    if adata.raw is None:
        adata.raw = adata

    # Copy of the raw data if it exisits
    if adata.raw is not None:
        adata.X = adata.raw.X

    # Make a copy of the data with the marker of interest
    data = pd.DataFrame(np.log1p(adata.X), columns = adata.var.index, index= adata.obs.index)[[marker_of_interest]]

    # Generate a dataframe with various gates
    def gate (g, d):
        dd = d.values
        dd = np.where(dd < g, np.nan, dd)
        np.warnings.filterwarnings('ignore')
        dd = np.where(dd > g, 1, dd)
        dd = pd.DataFrame(dd, index = d.index, columns = ['gate-' + str(g)])
        return dd

    # Identify the list of increments
    inc = list(np.arange (from_gate, to_gate, increment))
    inc = [round(num,3) for num in inc]

    # Apply the function
    r_gate = lambda x: gate(g=x, d=data) # Create lamda function
    gated_data = list(map(r_gate, inc)) # Apply function
    # Concat all the results into a single dataframe
    gates = pd.concat(gated_data, axis=1)

    ##########################################################################
    # Visulaisation using Napari
    # Plot only the Image that is requested
    if subset is not None:
        adata = adata[adata.obs[imageid] == subset]

    # Recover the channel names from adata
    if channel_names is 'default':
        channel_names = adata.uns['all_markers']
    else:
        channel_names = channel_names


    # Index of the marker of interest and corresponding names
    if markers is None:
        idx = [0, list(channel_names).index(marker_of_interest)]
        channel_names = [adata.uns['all_markers'][0], marker_of_interest]
    elif markers == 'all':
        idx = list(range(len(channel_names)))
        channel_names = channel_names
    else:
        markers = [marker_of_interest] + list(markers)
        idx = []
        for i in markers:
            idx.append(list(channel_names).index(i))
        channel_names = markers

    # Load the image
    image = tiff.imread(image_path, key = idx)

    # Load the segmentation mask
    if seg_mask is not None:
        seg_m = tiff.imread(seg_mask)

    # Load the viewer
    viewer = napari.view_image(
    image,
    #is_pyramid=False,
    channel_axis=0,
    name = None if channel_names is None else channel_names,
    visible = False)

    # Add the seg mask
    if seg_mask is not None:
        viewer.add_labels(seg_m,
                          name='segmentation mask',
                          visible = False)

    # subset the gates to include only the image of interest
    gates = gates.loc[adata.obs.index,]

    # Add phenotype layer function
    def add_phenotype_layer (adata, gates, phenotype_layer,x,y,viewer,point_size):
        cells = gates[gates[phenotype_layer] == 1].index
        coordinates = adata[cells]
        coordinates = pd.DataFrame({'y': coordinates.obs[y],'x': coordinates.obs[x]})
        points = coordinates.values.tolist()
        viewer.add_points(points, size=point_size, face_color='white',visible=False,name=phenotype_layer)

    # Run the function on all phenotypes
    for i in gates.columns:
        add_phenotype_layer (adata=adata, gates=gates, phenotype_layer=i, x=x_coordinate, y=y_coordinate, viewer=viewer,
                                                           point_size=point_size)



from ._phenotype_cells import phenotype_cells
from ._spatial_count import spatial_count
from ._spatial_expression import spatial_expression
from ._spatial_aggregate import spatial_aggregate
from ._cluster import cluster 
from ._spatial_interaction import spatial_interaction 
from ._spatial_distance import spatial_distance 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 15:00:39 2020
@author: Ajit Johnson Nirmal
Function to compute the proportion of phenotypes or any single cell annotation 
within the local neighbourhood of each cell. 
"""

# Import library
import pandas as pd
import numpy as np
from sklearn.neighbors import BallTree


# Function
def spatial_count (adata,x_coordinate='X_centroid',y_coordinate='Y_centroid',
                   phenotype='phenotype',method='radius',radius=30,knn=10,
                   imageid='imageid',subset=None,label='spatial_count'):
    """
    

    Parameters
    ----------
    adata : AnnData object

    x_coordinate : float, required
        Column name containing the x-coordinates values. The default is 'X_centroid'.
    y_coordinate : float, required
        Column name containing the y-coordinates values. The default is 'Y_centroid'.
    phenotype : string, required
        Column name of the column containing the phenotype information. 
        It could also be any categorical assignment given to single cells. The default is 'phenotype'.
    method : string, optional
        Two options are available: a) 'radius', b) 'knn'.
        a) radius - Identifies the neighbours within a given radius for every cell.
        b) knn - Identifies the K nearest neigbours for every cell.
        The default is 'radius'.
    radius : int, optional
        The radius used to define a local neighbhourhood. The default is 30.
    knn : int, optional
        Number of cells considered for defining the local neighbhourhood. The default is 10.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    subset : string, optional
        imageid of a single image to be subsetted for analyis. The default is None.
    label : string, optional
        Key for the returned data, stored in `adata.uns`. The default is 'spatial_count'.

    Returns
    -------
    adata : AnnData object
        Updated AnnData object with the results stored in `adata.uns['spatial_count']`.
    
    Example
    -------
    # Running the radius method
    adata = sm.tl.spatial_count (adata,x_coordinate='X_centroid',y_coordinate='Y_centroid',
                           phenotype='phenotype',method='radius',radius=30,
                           imageid='imageid',subset=None,label='spatial_count_radius')
    # Running the knn method
    adata = sm.tl.spatial_count (adata,x_coordinate='X_centroid',y_coordinate='Y_centroid',
                           phenotype='phenotype',method='knn',knn=10,
                           imageid='imageid',subset=None,label='spatial_count_knn')

    """

    def spatial_count_internal (adata_subset,x_coordinate,y_coordinate,phenotype,method,radius,knn,
                                imageid,subset,label):

        # Create a DataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate], 'phenotype': adata_subset.obs[phenotype]})
        
        # Identify neighbourhoods based on the method used
        # a) KNN method
        if method == 'knn':
            print("Identifying the " + str(knn) + " nearest neighbours for every cell")
            tree = BallTree(data[['x','y']], leaf_size= 2)
            ind = tree.query(data[['x','y']], k=knn, return_distance= False)
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            neighbours.drop(0, axis=1, inplace=True) # Remove self neighbour
        
        # b) Local radius method
        if method == 'radius':
            print("Identifying neighbours within " + str(radius) + " pixels of every cell")
            kdt = BallTree(data[['x','y']], metric='euclidean') 
            ind = kdt.query_radius(data[['x','y']], r=radius, return_distance=False)
            for i in range(0, len(ind)): ind[i] = np.delete(ind[i], np.argwhere(ind[i] == i))#remove self
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            
        # Map phenotype
        phenomap = dict(zip(list(range(len(ind))), data['phenotype'])) # Used for mapping
        
        # Loop through (all functionized methods were very slow)
        for i in neighbours.columns:
            neighbours[i] = neighbours[i].dropna().map(phenomap, na_action='ignore')
        
        # Drop NA
        #n_dropped = neighbours.dropna(how='all')
           
        # Collapse all the neighbours into a single column
        n = pd.DataFrame(neighbours.stack(), columns = ["neighbour_phenotype"])
        n.index = n.index.get_level_values(0) # Drop the multi index
        n = pd.DataFrame(n)
        n['order'] = list(range(len(n)))
        
        # Merge with real phenotype
        n_m = n.merge(data['phenotype'], how='inner', left_index=True, right_index=True)
        n_m['neighbourhood'] = n_m.index
        n = n_m.sort_values(by=['order'])
        
        # Normalize based on total cell count
        k = n.groupby(['neighbourhood','neighbour_phenotype']).size().unstack().fillna(0)
        k = k.div(k.sum(axis=1), axis=0)
        
        # return the normalized neighbour occurance count
        return k
    
    # Subset a particular image if needed
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
    
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_count_internal = lambda x: spatial_count_internal(adata_subset=x,x_coordinate=x_coordinate,
                                                   y_coordinate=y_coordinate,phenotype=phenotype,
                                                   method=method,radius=radius,knn=knn,
                                                   imageid=imageid,subset=subset,label=label) 
    all_data = list(map(r_spatial_count_internal, adata_list)) # Apply function 
    
    
    # Merge all the results into a single dataframe    
    result = []
    for i in range(len(all_data)):
        result.append(all_data[i])
    result = pd.concat(result, join='outer')  
    
    # Reindex the cells
    result = result.fillna(0)
    result = result.reindex(adata.obs.index)
    
    # Add to adata
    adata.uns[label] = result
    
    # Return        
    return adata


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 12 23:47:52 2020
@author: Ajit Johnson Nirmal
View the gating in Napari viewer
"""

import napari
import pandas as pd
import tifffile as tiff
import numpy as np

import dask.array as da
import zarr
import pathlib

def gate_finder (image_path, adata, marker_of_interest, from_gate = 6, to_gate = 8, increment = 0.1,
                 markers=None, channel_names = 'default',
                 x_coordinate='X_centroid',y_coordinate='Y_centroid',
                 point_size=10,imageid='imageid',subset=None,seg_mask=None,**kwargs):
    """


    Parameters
    ----------
    image_path : string
        Location to the image file..
    adata : Ann Data Object
    marker_of_interest : string
        Marker for which gate is to be defined e.g. 'CD45'.
    from_gate : int, optional (The default is 6)
        Start value gate of interest.
    to_gate : int, optional (The default is 8)
        End value of the gate of interest.
    increment : float, optional (The default is 0.1)
        Increments between the start and end values.
    markers : string, optional (The default is None)
        Additional markers to be included in the plot for evaluation.
    channel_names : list, optional (The default is `adata.uns['all_markers']`)
        List of channels in the image in the exact order as image.
    x_coordinate : string, optional (The default is 'X_centroid')
        X axis coordinate column name in AnnData object.
    y_coordinate : string, optional (The default is 'Y_centroid')
        Y axis coordinate column name in AnnData object.
    point_size : int, optional (The default is 10)
        point size in the napari plot.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    subset : string, optional
        imageid of a single image to be subsetted for analyis. The default is None.
    seg_mask : string, optional (The default is None)
        Location to the segmentation mask file.
    **kwargs
        Other arguments that can be passed to napari viewer

    Example
    -------
    image_path = '/Users/aj/Desktop/ptcl_tma/image.tif'
    sm.pl.gate_finder (image_path, adata, marker_of_interest='CD45',
                 from_gate = 6, to_gate = 8, increment = 0.1,
                 markers=['DNA10'], channel_names = 'default',
                 x_coordinate='X_position',y_coordinate='Y_position',point_size=10,
                 subset= '77', seg_mask=None)

    """

    # If no raw data is available make a copy
    if adata.raw is None:
        adata.raw = adata

    # Copy of the raw data if it exisits
    if adata.raw is not None:
        adata.X = adata.raw.X

    # Make a copy of the data with the marker of interest
    data = pd.DataFrame(np.log1p(adata.X), columns = adata.var.index, index= adata.obs.index)[[marker_of_interest]]

    # Generate a dataframe with various gates
    def gate (g, d):
        dd = d.values
        dd = np.where(dd < g, np.nan, dd)
        np.warnings.filterwarnings('ignore')
        dd = np.where(dd > g, 1, dd)
        dd = pd.DataFrame(dd, index = d.index, columns = ['gate-' + str(g)])
        return dd

    # Identify the list of increments
    inc = list(np.arange (from_gate, to_gate, increment))
    inc = [round(num,3) for num in inc]

    # Apply the function
    r_gate = lambda x: gate(g=x, d=data) # Create lamda function
    gated_data = list(map(r_gate, inc)) # Apply function
    # Concat all the results into a single dataframe
    gates = pd.concat(gated_data, axis=1)

    ##########################################################################
    # Visulaisation using Napari
    
    # Load the image    
    image = tiff.TiffFile(image_path, is_ome=False)
    z = zarr.open(image.aszarr(), mode='r') # convert image to Zarr array
    #z = image.aszarr() # convert image to Zarr array
    
    # Plot only the Image that is requested
    if subset is not None:
        adata = adata[adata.obs[imageid] == subset]

    # Recover the channel names from adata
    if channel_names is 'default':
        channel_names = adata.uns['all_markers']
    else:
        channel_names = channel_names
        
    # if markers is a string convert to list
    if isinstance(markers, str):
        markers = [markers]
        
    # Index of the marker of interest and corresponding names
    if markers is not None:
        markers.extend([marker_of_interest])
        idx = np.where(np.isin(channel_names,markers))[0]
        channel_names = [channel_names[i] for i in idx]
    else:
        idx = list(range(len(channel_names)))
        channel_names = channel_names
        

    # Identify the number of pyramids and number of channels
    n_levels = len(image.series[0].levels) # pyramid
    
    # If and if not pyramids are available
    if n_levels > 1:
        pyramid = [da.from_zarr(z[i]) for i in range(n_levels)]
        multiscale = True
    else:
        pyramid = da.from_zarr(z)
        multiscale = False
    
    # subset channels of interest
    if markers is not None:
        if n_levels > 1:
            for i in range(n_levels-1):
                pyramid[i] = pyramid[i][idx, :, :]
            n_channels = pyramid[0].shape[0] # identify the number of channels
        else:
            pyramid = pyramid[idx, :, :]
            n_channels = pyramid.shape[0] # identify the number of channels
    else:
        if n_levels > 1:
            n_channels = pyramid[0].shape[0]
        else:
            n_channels = pyramid.shape[0]
            
    
    # check if channel names have been passed to all channels
    if channel_names is not None:
        assert n_channels == len(channel_names), (
            f'number of channel names ({len(channel_names)}) must '
            f'match number of channels ({n_channels})'
        )
            
        

    # Load the segmentation mask
    if seg_mask is not None:
        seg_m = tiff.imread(seg_mask)


    # Load the viewer
    viewer = napari.view_image(
    pyramid,
    multiscale=multiscale,
    channel_axis=0,
    name = None if channel_names is None else channel_names,
    visible = False, **kwargs)

    # Add the seg mask
    if seg_mask is not None:
        viewer.add_labels(seg_m,
                          name='segmentation mask',
                          visible = False)

    # subset the gates to include only the image of interest
    gates = gates.loc[adata.obs.index,]

    # Add phenotype layer function
    def add_phenotype_layer (adata, gates, phenotype_layer,x,y,viewer,point_size):
        cells = gates[gates[phenotype_layer] == 1].index
        coordinates = adata[cells]
        coordinates = pd.DataFrame({'y': coordinates.obs[y],'x': coordinates.obs[x]})
        points = coordinates.values.tolist()
        viewer.add_points(points, size=point_size, face_color='white',visible=False,name=phenotype_layer)

    # Run the function on all phenotypes
    for i in gates.columns:
        add_phenotype_layer (adata=adata, gates=gates, phenotype_layer=i, x=x_coordinate, y=y_coordinate, viewer=viewer,
                                                           point_size=point_size)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 12 19:06:40 2020
@author: Ajit Johnson Nirmal
Manual Gate finder
"""

import pandas as pd
import numpy as np
from sklearn.mixture import GaussianMixture 
from scipy import stats
import seaborn as sns; sns.set(color_codes=True)
import matplotlib.pyplot as plt



def gmm_gate (adata, marker_of_interest):
    """
    

    Parameters
    ----------
    adata : Ann Data Object
    marker_of_interest : string
        Marker of interest.

    Returns
    -------
    Distribution plot of the marker of interest along with GMM overlaid.
    
    Example
    -------
    sm.pl.gmm_dist_plot (adata, marker_of_interest='CD45')

    """
    
    # If no raw data is available make a copy
    if adata.raw is None:
        adata.raw = adata
    
    # Copy of the raw data if it exisits
    if adata.raw is not None:
        adata.X = adata.raw.X
    
    # Make a copy of the data with the marker of interest
    data = pd.DataFrame(np.log1p(adata.X), columns = adata.var.index, index= adata.obs.index)[[marker_of_interest]]
    
    # Clip of the top and bottom outliers before applying the model
    data = data.clip(lower =np.percentile(data,1), upper=np.percentile(data,99))
    
    # Apply Gaussian mixture model
    m = data[marker_of_interest].values
    data_gm = m.reshape(-1, 1)
    gmm = GaussianMixture(n_components=2)
    gmm.fit(data_gm)
    gate = np.mean(gmm.means_)
    
    mean = gmm.means_
    cov = gmm.covariances_
    std = [ np.sqrt(  np.trace(cov[i])/2) for i in range(0,2) ]
    #std = np.sqrt(np.trace(cov)/2) 
    
    
    # Merge with image ID
    if len(adata.obs['ImageId'].unique()) > 1:
        # Identify the most and least positivity of the given marker of interest
        # Generate a dataframe with various gates
        dd = data.values
        dd = np.where(dd < gate, np.nan, dd)
        np.warnings.filterwarnings('ignore')
        dd = np.where(dd > gate, 1, dd)
        dd = pd.DataFrame(dd, index = data.index, columns = ['gate'])
        dd = dd.merge(pd.DataFrame(adata.obs['ImageId']), how='outer', left_index=True, right_index=True)
        dd = dd[dd['gate'] == 1]
        image_positivity = dd.groupby('ImageId').count().sort_values('gate')
        # Create a string for overlaying on image
        low = list(image_positivity.head(3).index)
        low = ','.join(low)
        high = list(image_positivity.tail(3).index)
        high = ','.join(high)
        # Dor image
        textstr = '\n'.join((
            ("Most positive in: ", 
             high,
             "Least positive in ",
             low
            )))

    # Plot
    sns.set_style("white")
    fig, ax = plt.subplots()
    x = np.linspace(mean[0] - 3*std[0], mean[0] + 3*std[0], 1000)
    y = np.linspace(mean[1] - 3*std[1], mean[1] + 3*std[1], 1000)
    sns.distplot(data[marker_of_interest], color = 'grey')
    plt.axvline(x= mean[0], c='#000000', linestyle='dotted')
    plt.axvline(x= mean[1], c='#000000', linestyle='dotted')
    plt.plot(x, stats.norm.pdf(x, mean[0], std[0]), linestyle='dashed')
    plt.plot(y, stats.norm.pdf(y, mean[1], std[1]), linestyle='dashed')
    plt.xticks(np.arange(min(data[marker_of_interest])-1, max(data[marker_of_interest])+1, 0.5))
    plt.title(marker_of_interest, fontsize=20)    
    if len(adata.obs['ImageId'].unique()) > 1:
        ax.text(0.05, 0.95, textstr, fontsize=14, transform=ax.transAxes,
            verticalalignment='top')





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 14 09:23:07 2020
@author: Ajit Johnson Nirmal
This function will allow users to calculate the average shortest between every cell to phenotypes/clusters
of interest.  
"""

# Import library
import pandas as pd
from sklearn.neighbors import BallTree
from joblib import Parallel, delayed
import itertools

# Function

def spatial_distance (adata,x_coordinate='X_centroid',y_coordinate='Y_centroid',
                      phenotype='phenotype',subset=None,imageid='imageid',
                      label='spatial_distance'):
    """
    

    Parameters
    ----------
    adata : AnnData object

    x_coordinate : float, required
        Column name containing the x-coordinates values. The default is 'X_centroid'.
    y_coordinate : float, required
        Column name containing the y-coordinates values. The default is 'Y_centroid'.
    phenotype : string, required
        Column name of the column containing the phenotype information. 
        It could also be any categorical assignment given to single cells. The default is 'phenotype'.
    subset : string, optional
        imageid of a single image to be subsetted for analyis. The default is None.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    label : string, optional
        Key for the returned data, stored in `adata.obs`. The default is 'spatial_distance'.

    Returns
    -------
    adata : AnnData object
        Updated AnnData object with the results stored in `adata.uns['spatial_distance']`.
        
    Example
    -------
    adata = sm.tl.spatial_distance (adata,x_coordinate='X_position',y_coordinate='Y_position',imageid='ImageId')
        

    """
    
    
    def spatial_distance_internal (adata_subset,x_coordinate,y_coordinate,
                                   phenotype,subset,imageid,label):
        
        print("Processing Image: " + str(adata_subset.obs[imageid].unique()[0]))
        # Create a dataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate], 'phenotype': adata_subset.obs[phenotype]})

        # Function to identify shortest distance for each phenotype of interest
        def distance (pheno):
            pheno_interest = data[data['phenotype'] == pheno]
            # Build the ball-tree for search space
            tree = BallTree(pheno_interest[['x','y']], metric='euclidean') 
            # Calculate shortest distance
            dist, ind = tree.query(data[['x','y']], k=1, return_distance= True)
            dist = list(itertools.chain.from_iterable(dist))
            return dist
        
        # Run in parallel for all phenotypes
        phenotype_list = list(data['phenotype'].unique())
        # Apply function
        final_dist = Parallel(n_jobs=-1)(delayed(distance)(pheno=i) for i in phenotype_list)     
        final_dist = pd.DataFrame(final_dist, index = phenotype_list, columns = data.index).T
        
        return final_dist
    
    # subset a particular subset of cells if the user wants else break the adata into list of anndata objects
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
        
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_distance_internal = lambda x: spatial_distance_internal (adata_subset=x,x_coordinate=x_coordinate,y_coordinate=y_coordinate, phenotype=phenotype,subset=subset,imageid=imageid,label=label) 
    all_data = list(map(r_spatial_distance_internal, adata_list)) # Apply function 
    
    # Merge all the results into a single dataframe    
    result = []
    for i in range(len(all_data)):
        result.append(all_data[i])
    result = pd.concat(result, join='outer')  
    
    
    # Add to anndata
    adata.uns[label] = result
    
    # return
    return adata
    
        




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 10 09:32:49 2020
@author: Ajit Johnson Nirmal
Tests
"""

import pytest
import sys, os

#os.chdir ("/Users/aj/Dropbox (Partners HealthCare)/packages/scimap")

@pytest.fixture
def adata():
    import anndata as ad    
    adata = ad.read(os. getcwd() + '/scimap/tests/_data/example_data.h5ad')
    return adata

# test rename function 
def test_rename (adata):
    from scimap.helpers._rename import rename
    name= {'T cells': ['CD8 T cells', 'CD4 T cells']}
    adata = rename (adata, name, from_column='phenotype', to_column='phenotype_renamed')
    assert adata.obs['phenotype_renamed'].value_counts()['T cells'] == 97
    

# test classify function
def test_classify (adata):
    from scimap.helpers._classify import classify
    adata = classify(adata, pos=['CD3D','FOXP3'], neg=['ASMA'], 
                     subclassify_phenotype=['T cells','Regulatory T cells'])
    
    a = adata.obs['classify'].value_counts()['passed_classify']
    assert a == 16
    

# test scimap_to_csv function
def test_scimap_to_csv (adata):
    from scimap.helpers._scimap_to_csv import scimap_to_csv
    data = scimap_to_csv(adata)
    
    # test
    a = data.shape[1]
    assert a == 45


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 12 17:03:56 2020
@author: Ajit Johnson Nirmal
Function to sub-cluster a cluster of interest. Particularly useful to check if 
there are sub-phenotypes after performing the gating based phenotyping.
"""

# Import library
import numpy as np
import pandas as pd
import scanpy as sc
import scanpy.external as sce
from sklearn.cluster import KMeans
try:
    import parc
except:
    pass



def cluster (adata, method = 'kmeans', subset_genes=None,
             sub_cluster=False, sub_cluster_column='phenotype', sub_cluster_group = None,
             parc_small_pop= 50, parc_too_big_factor=0.4, 
             k= 10, n_pcs=None, resolution=1, 
             phenograph_clustering_metric='euclidean', nearest_neighbors= 30, 
             use_raw = True, random_state=0, collapse_labels= False,
             label=None):
    """
    
    Parameters
    ----------
    adata : AnnData Object
    method : string, optional
        Clustering method to be used- Implemented methods- kmeans, phenograph, leiden and parc. The default is 'kmeans'.
    subset_genes : list, optional
        Pass a list of genes ['CD3D', 'CD20', 'KI67'] that should be included for the purpose of clustering. 
        By default the algorithm uses all genes in the dataset.
    sub_cluster : Boolian, optional
        If the user has already performed clustering or phenotyping previously and would like to
        sub-cluster within a particular cluster/phenotype, this option can be used. The default is False.
    sub_cluster_column : string, optional
        The column name that contains the cluster/phenotype information to be sub-clustered. 
        This is only required when sub_cluster is set to True.
        The default is 'phenotype'.
    sub_cluster_group : list, optional
        By default the program will sub-cluster all groups within column passed through the argument sub_cluster_column.
        If user wants to sub cluster only a subset of phenotypes/clusters this option can be used.
        Pass them as list e.g. ["tumor", "b cells"].     
    parc_small_pop : int, optional
        Smallest cluster population to be considered a community in PARC clustering. The default is 50.
    parc_too_big_factor : float, optional
        If a cluster exceeds this share of the entire cell population, then the PARC will be run on 
        the large cluster. at 0.4 it does not come into play. The default is 0.4.
    k : int, optional
        Number of clusters to return when using K-Means clustering. The default is 10.
    n_pcs : int, optional
        Number of PC's to be used in leiden clustering. By default it uses all PC's.
    resolution : float, optional
        A parameter value controlling the coarseness of the clustering. 
        Higher values lead to more clusters. The default is 1.
    phenograph_clustering_metric : string, optional
        Distance metric to define nearest neighbors. Note that performance will be slower for correlation and cosine. 
        Available methods- cityblock’, ‘cosine’, ‘euclidean’, ‘manhattan’, braycurtis’, ‘canberra’, ‘chebyshev’, 
        ‘correlation’, ‘dice’, ‘hamming’, ‘jaccard’, ‘kulsinski’, ‘mahalanobis’, ‘minkowski’, ‘rogerstanimoto’, 
        ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’
        The default is 'euclidean'.
    nearest_neighbors : int, optional
        Number of nearest neighbors to use in first step of graph construction. 
        This parameter is used both in leiden and phenograph clustering.
        The default is 30.
    use_raw : bool, optional
        If True, log transformed raw data will be used for clustering. 
        If False, normalized/scaled data within `adata.X` will be used. The default is True.
    random_state : int, optional
        Change the initialization of the optimization. The default is 0.
    collapse_labels : bool, optional
        While sub clustering only a few phenotypes/clusters, this argument helps to 
        group all the other phenotypes/clusters into a single category- 
        Helps in visualisation. The default is False.
    label : string, optional
        Key or optional column name for the returned data, stored in `adata.obs`. The default is adata.obs[method used].


    Returns
    -------
    adata : AnnData Object
        Returns an updated anndata object with a new column. check- adata.obs[method used]
        
    Example
    -------
    adata = sm.tl.cluster (adata, k= 10, method = 'kmeans', sub_cluster_column='phenotype', use_raw = True)


    """
    
    
    # Leiden clustering
    def leiden_clustering (pheno, adata, nearest_neighbors, n_pcs, resolution):
        
        # subset the data to be clustered
        if pheno is not None:
            cell_subset =  adata.obs[adata.obs[sub_cluster_column] == pheno].index
        else:
            cell_subset = adata.obs.index
        
        if use_raw == True:
            data_subset = adata[cell_subset]
            data_subset.X = np.log1p(data_subset.raw.X)          
        else:
            data_subset = adata[cell_subset]
        
        # clustering
        if pheno is not None:
            print('Leiden clustering ' + str(pheno))
        else:
            print('Leiden clustering')
            
        sc.tl.pca(data_subset)
        if n_pcs is None:
            n_pcs = len(adata.var)
        sc.pp.neighbors(data_subset, n_neighbors=nearest_neighbors, n_pcs=n_pcs)
        sc.tl.leiden(data_subset,resolution=resolution, random_state=random_state)
        
        # Rename the labels
        cluster_labels = list(map(str,list(data_subset.obs['leiden'])))
        if pheno is not None:
            cluster_labels = list(map(lambda orig_string: pheno + '-' + orig_string, cluster_labels))

        # Make it into a dataframe
        cluster_labels = pd.DataFrame(cluster_labels, index = data_subset.obs.index)
        
        # return labels
        return cluster_labels

    # Kmeans clustering
    def k_clustering (pheno, adata, k, sub_cluster_column, use_raw, random_state):
        
        # subset the data to be clustered
        if pheno is not None:
            cell_subset =  adata.obs[adata.obs[sub_cluster_column] == pheno].index
        else:
            cell_subset = adata.obs.index
        
        # Usage of scaled or raw data
        if use_raw == True:
            data_subset = pd.DataFrame(np.log1p(adata.raw[cell_subset].X), columns =adata[cell_subset].var.index, index = adata[cell_subset].obs.index)
        else:
            data_subset = pd.DataFrame(adata[cell_subset].X, columns =adata[cell_subset].var.index, index = adata[cell_subset].obs.index)
            
        # K-means clustering
        if pheno is not None:
            print('Kmeans clustering ' + str(pheno))
        else:
            print('Kmeans clustering')
        
        kmeans = KMeans(n_clusters=k, random_state=random_state).fit(data_subset)
        
        # Rename the labels
        cluster_labels = list(map(str,kmeans.labels_))
        if pheno is not None:
            cluster_labels = list(map(lambda orig_string: pheno + '-' + orig_string, cluster_labels))
                
        # Make it into a 
        cluster_labels = pd.DataFrame(cluster_labels, index = data_subset.index)
        
        # return labels
        return cluster_labels
        
    # Phenograph clustering
    def phenograph_clustering (pheno, adata, primary_metric, nearest_neighbors):
        
        # subset the data to be clustered
        if pheno is not None:
            cell_subset =  adata.obs[adata.obs[sub_cluster_column] == pheno].index
        else:
            cell_subset = adata.obs.index
        
        # Usage of scaled or raw data
        if use_raw == True:
            data_subset = adata[cell_subset]
            data_subset.X = np.log1p(data_subset.raw.X)          
        else:
            data_subset = adata[cell_subset]
        
        # Phenograph clustering
        if pheno is not None:
            print('Phenograph clustering ' + str(pheno))
        else:
            print('Phenograph clustering')
        
        sc.tl.pca(data_subset)
        result = sce.tl.phenograph(data_subset.obsm['X_pca'], k = nearest_neighbors, primary_metric=phenograph_clustering_metric)
        
        # Rename the labels
        cluster_labels = list(map(str,result[0]))
        if pheno is not None:
            cluster_labels = list(map(lambda orig_string: pheno + '-' + orig_string, cluster_labels))
        
        # Make it into a dataframe
        cluster_labels = pd.DataFrame(cluster_labels, index = data_subset.obs.index)
        
        # return labels
        return cluster_labels
        
    
    # PARC clustering
    # https://github.com/ShobiStassen/PARC
    def parc_clustering (pheno, adata, random_state,resolution,parc_too_big_factor,parc_small_pop):
        
        # subset the data to be clustered
        if pheno is not None:
            cell_subset =  adata.obs[adata.obs[sub_cluster_column] == pheno].index
        else:
            cell_subset = adata.obs.index
        
        # Usage of scaled or raw data
        if use_raw == True:
            data_subset = adata[cell_subset]
            data_subset.X = np.log1p(data_subset.raw.X)          
        else:
            data_subset = adata[cell_subset]
        
        # Phenograph clustering
        if pheno is not None:
            print('Parc clustering ' + str(pheno))
        else:
            print('Parc clustering')
        
        sc.tl.pca(data_subset)
        parc1 = parc.PARC(data_subset.obsm['X_pca'], random_seed=random_state, parc_small_pop = parc_small_pop, resolution_parameter=resolution,parc_too_big_factor=parc_too_big_factor)  
        parc1.run_PARC() # Run Parc
        

        # Rename the labels
        cluster_labels = list(map(str,parc1.labels))
        if pheno is not None:
            cluster_labels = list(map(lambda orig_string: pheno + '-' + orig_string, cluster_labels))
        
        # Make it into a dataframe
        cluster_labels = pd.DataFrame(cluster_labels, index = data_subset.obs.index)
        
        # return labels
        return cluster_labels
    
    # Use user defined genes for clustering
    if subset_genes is not None:
        bdata = adata[:,subset_genes]
        bdata.raw = bdata[:,subset_genes]
    else:
        bdata = adata
        
    # What cells to run the clustering on?
    if sub_cluster_group is not None:
        if isinstance(sub_cluster_group, list):
            pheno = sub_cluster_group
        else:
            pheno = [sub_cluster_group]         
    else:
        # Make sure number of clusters is not greater than number of cells available
        if method == 'kmeans':
            pheno = (bdata.obs[sub_cluster_column].value_counts() > k+1).index[bdata.obs[sub_cluster_column].value_counts() > k+1]
        if method == 'phenograph':
            pheno = (bdata.obs[sub_cluster_column].value_counts() > nearest_neighbors+1).index[bdata.obs[sub_cluster_column].value_counts() > nearest_neighbors+1]
        if method == 'leiden':
            pheno = (bdata.obs[sub_cluster_column].value_counts() > 1).index[bdata.obs[sub_cluster_column].value_counts() > 1]
        if method == 'parc':
            pheno = (bdata.obs[sub_cluster_column].value_counts() > 1).index[bdata.obs[sub_cluster_column].value_counts() > 1]
            
        
    # Run the specified method
    if method == 'kmeans':
        if sub_cluster == True:  
            # Apply the Kmeans function
            r_k_clustering = lambda x: k_clustering(pheno=x, adata=bdata, k=k, sub_cluster_column=sub_cluster_column, use_raw=use_raw, random_state=random_state) # Create lamda function 
            all_cluster_labels = list(map(r_k_clustering, pheno)) # Apply function 
        else:
            all_cluster_labels = k_clustering(pheno=None, adata=bdata, k=k, sub_cluster_column=sub_cluster_column, use_raw=use_raw, random_state=random_state)
            
    if method == 'phenograph':
        if sub_cluster == True:
            r_phenograph_clustering = lambda x: phenograph_clustering(pheno=x, adata=bdata, primary_metric=phenograph_clustering_metric, nearest_neighbors=nearest_neighbors) # Create lamda function 
            all_cluster_labels = list(map(r_phenograph_clustering, pheno)) # Apply function      
        else:
            all_cluster_labels = phenograph_clustering(pheno=None, adata=bdata, primary_metric=phenograph_clustering_metric, nearest_neighbors=nearest_neighbors)
            
            
    if method == 'leiden':
        if sub_cluster == True:
            r_leiden_clustering = lambda x: leiden_clustering(pheno=x, adata=bdata, nearest_neighbors=nearest_neighbors, n_pcs=n_pcs, resolution=resolution) # Create lamda function 
            all_cluster_labels = list(map(r_leiden_clustering, pheno)) # Apply function 
        else:
            all_cluster_labels = leiden_clustering(pheno=None, adata=bdata, nearest_neighbors=nearest_neighbors, n_pcs=n_pcs, resolution=resolution)
            
            
    if method == 'parc':
        if sub_cluster == True:
            r_parc_clustering = lambda x: parc_clustering(pheno=x, adata=bdata, random_state=random_state,resolution=resolution,parc_too_big_factor=parc_too_big_factor,parc_small_pop=parc_small_pop) # Create lamda function 
            all_cluster_labels = list(map(r_parc_clustering, pheno)) # Apply function 
        else:
            all_cluster_labels = parc_clustering(pheno=None, adata=bdata, random_state=random_state,resolution=resolution,parc_too_big_factor=parc_too_big_factor,parc_small_pop=parc_small_pop)
       
    
    # Merge all the labels into one and add to adata
    if sub_cluster == True:
        sub_clusters = pd.concat(all_cluster_labels, axis=0, sort=False)
    else:
        sub_clusters = all_cluster_labels
        
    # Merge with all cells
    sub_clusters = pd.DataFrame(bdata.obs[sub_cluster_column]).merge(sub_clusters, how='outer', left_index=True, right_index=True)
        

    
    # Transfer labels
    if collapse_labels == False:
        sub_clusters = pd.DataFrame(sub_clusters[0].fillna(sub_clusters[sub_cluster_column]))
        
    # Get only the required column
    sub_clusters = sub_clusters[0]
    
    # re index the rows
    sub_clusters = sub_clusters.reindex(adata.obs.index)
    
    # Append to adata
    if label is None:
        adata.obs[method] = sub_clusters
    else:
        adata.obs[label] = sub_clusters

    
    # Return adata
    return adata



# -*- coding: utf-8 -*-
"""
Created on Mon Mar  2 09:12:35 2020
@author: Ajit Johnson Nirmal
Convert mcmicro output to AnnData object
"""

# Import library
import numpy as np
import anndata as ad
import pandas as pd

def mcmicro_to_scimap (image_path,remove_dna=True,remove_string_from_name=None,
                        log=True,drop_markers=None,random_sample=None, unique_CellId=True,
                        CellId='CellID',split='X_centroid',custom_imageid=None,
                        min_cells=None):
    """

    Parameters
    ----------
    image_path : list
        List of path to the image or images. Each Image should have a unique path supplied.
    remove_dna : bool, optional
        Remove the DNA channels from the final output. Looks for channels with the string 'dna' in it. The default is True.
    remove_string_from_name : string, optional
        Used to celan up channel names. If a string is given, that particular string will be removed from all marker names.
        If multiple images are passed, just use the string that appears in the first image. The default is None.
    log : bool, optional
        Log the data (log1p transformation will be applied). The default is True.
    drop_markers : list, optional
        List of markers to drop from the analysis. e.g. ["CD3D", "CD20"]. The default is None.
    random_sample : int, optional
        Randomly sub-sample the data with the desired number of cells. The default is None.
    CellId : string, optional
        Name of the column that contains the cell ID. The default is CellID.
    unique_CellId: bool, optional
        By default, the function creates a unique name for each cell/row by combining the 
        `CellId` and `imageid`. If you wish not to perform this operation please pass `False`.
        The function will use whatever is under `CellId`. In which case, please be careful to pass unique `CellId`
        especially when loading multiple datasets togeather.  
    split : string, optional
        To split the CSV into counts table and meta data, pass in the name of the column
        that immediately follows the marker quantification. The default is 'X_centroid'.
    custom_imageid: string, optional
        Pass a user defined Image ID. By default the name of the CSV file is used.
    min_cells: int, optional
        If these many cells are not in the image, the image will be dropped.
        Particulary useful when importing multiple images.

    Returns
    -------
    AnnData Object

    Example
    -------
    image_path = ['/Users/aj/whole_sections/PTCL1_450.csv',
                  '/Users/aj/whole_sections/PTCL2_552.csv']
    adata = sm.pp.mcmicro_to_scimap (image_path, drop_markers= ['CD21', 'ACTIN'], random_sample=5000)

    """

    # Import data based on the location provided
    def load_process_data (image):
        # Print the data that is being processed
        print("Loading " + str(image.rsplit('/', 1)[-1]))
        d = pd.read_csv(image)
        # If the data does not have a unique image ID column, add one.
        if 'imageid' not in d.columns:
            if custom_imageid is not None:
                imid = custom_imageid
            else:
                #imid = random.randint(1000000,9999999)
                imid = str(image.rsplit('/', 1)[-1]).replace('.csv','')
            d['imageid'] = imid
        # Unique name for the data
        if unique_CellId is True:
            d.index = d['imageid'].astype(str)+'_'+d[CellId].astype(str)
        else:
            d.index = d[CellId]
        # Drop imageid and cell ID column
        d.drop([CellId], axis=1, inplace=True)
        # Move Image ID to the last column
        cols = d.columns.tolist()
        cols.insert(len(cols), cols.pop(cols.index('imageid')))
        d = d.reindex(columns= cols)
        # If there is INF replace with zero
        d = d.replace([np.inf, -np.inf], 0)
        # Return data
        return d
    # Apply function to all images and create a master dataframe
    r_load_process_data = lambda x: load_process_data(image=x) # Create lamda function
    all_data = list(map(r_load_process_data, list(image_path))) # Apply function

    # Merge all the data into a single large dataframe
    for i in range(len(all_data)):
        all_data[i].columns = all_data[0].columns
    entire_data = pd.concat(all_data, axis=0, sort=False)
    
    # Randomly sample the data
    if random_sample is not None:
        entire_data = entire_data.sample(n=random_sample,replace=False)

    #Remove the images that contain less than a defined threshold of cells (min_cells)
    if min_cells is not None:
        to_drop = entire_data['imageid'].value_counts()[entire_data['imageid'].value_counts() < min_cells].index
        entire_data = entire_data[~entire_data['imageid'].isin(to_drop)]
        print('Removed Images that contained less than '+str(min_cells)+' cells: '+ str(to_drop.values))

    # Split the data into expression data and meta data
    # Step-1 (Find the index of the column with name Area)
    split_idx = entire_data.columns.get_loc(split)
    meta = entire_data.iloc [:,split_idx:]
    # Step-2 (select only the expression values)
    entire_data = entire_data.iloc [:,:split_idx]

    # Rename the columns of the data
    if remove_string_from_name is not None:
        entire_data.columns = entire_data.columns.str.replace(remove_string_from_name, '')

    # Save a copy of the column names in the uns space of ANNDATA
    markers = list(entire_data.columns)

    # Remove DNA channels
    if remove_dna is True:
        entire_data = entire_data.loc[:,~entire_data.columns.str.contains('dna', case=False)]

    # Drop unnecessary markers
    if drop_markers is not None:
        for i in drop_markers:
            entire_data = entire_data.loc[:,~entire_data.columns.str.contains(i, case=False)]

    # Create an anndata object
    adata = ad.AnnData(entire_data)
    adata.obs = meta
    adata.uns['all_markers'] = markers

    # Add log data
    if log is True:
        adata.raw = adata
        adata.X = np.log1p(adata.X)


    # Return data
    return adata



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 15:00:39 2020
@author: Ajit Johnson Nirmal
Function to compute a proximity based weighted expression scoring. The function generates
a neighbourhood for each cell and computes a expression score for all markers
for each cell based on its proximity to other cells within the neighbourhood.
"""

# Import library
import pandas as pd
import numpy as np
from sklearn.neighbors import BallTree
from scipy.sparse import csr_matrix

# Function
def spatial_expression (adata, x_coordinate='X_centroid',y_coordinate='Y_centroid',
                        method='radius', radius=30, knn=10, imageid='imageid', 
                        use_raw=True,subset=None,label='spatial_expression'):
    """
    

    Parameters
    ----------
    adata : AnnData object

    x_coordinate : float, required
        Column name containing the x-coordinates values. The default is 'X_centroid'.
    y_coordinate : float, required
        Column name containing the y-coordinates values. The default is 'Y_centroid'.
    method : string, optional
        Two options are available: a) 'radius', b) 'knn'.
        a) radius - Identifies the neighbours within a given radius for every cell.
        b) knn - Identifies the K nearest neigbours for every cell.
        The default is 'radius'.
    radius : int, optional
        The radius used to define a local neighbhourhood. The default is 30.
    knn : int, optional
        Number of cells considered for defining the local neighbhourhood. The default is 10.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    subset : string, optional
        imageid of a single image to be subsetted for analyis. The default is None.
    use_raw : boolian, optional
        Argument to denote whether to use the raw data or scaled data after applying `sm.pp.rescale`.
        If `True`, the log of raw data is used. The default is True.
    label : string, optional
        Key for the returned data, stored in `adata.uns`. The default is 'spatial_count'.

    Returns
    -------
    adata : AnnData object
        Updated AnnData object with the results stored in `adata.uns['spatial_expression']`.
        
        
    Example
    -------
    # Running the radius method
    adata = sm.tl.spatial_expression (adata, x_coordinate='X_centroid',y_coordinate='Y_centroid',
                                method='radius', radius=30, imageid='imageid', 
                                use_raw=True,subset=None,label='spatial_expression_radius')
    # Running the knn method
    adata = sm.tl.spatial_expression (adata, x_coordinate='X_centroid',y_coordinate='Y_centroid',
                        method='knn', knn=10, imageid='imageid', 
                        use_raw=True,subset=None,label='spatial_expression_knn')

    """
    
    # Error statements
    if use_raw is False:
        if all(adata.X[0] < 1) is False:
            raise ValueError('Please run `sm.pp.rescale` first if you wish to use `use_raw = False`')
        
     
    def spatial_expression_internal (adata_subset, x_coordinate, y_coordinate,
                                     method, radius, knn, imageid, use_raw, subset,label):
         
        # Create a DataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate]})
        
        # Identify neighbourhoods based on the method used
        # a) KNN method
        if method == 'knn':
            print("Identifying the " + str(knn) + " nearest neighbours for every cell")
            tree = BallTree(data, leaf_size= 2)
            dist, ind = tree.query(data, k=knn, return_distance= True)
            # Remove self from ind
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            neighbours.drop(0, axis=1, inplace=True) # Remove self neighbour
            # Remove self from dist
            neighbours_dist = pd.DataFrame(dist.tolist(), index = data.index) # neighbour DF
            neighbours_dist.drop(0, axis=1, inplace=True) # Remove self neighbour
            # Merge the indeces and distance into a single dictionary
            ind_dict = lambda x,y: dict(zip(x, y)) # Function to create a dict of indeces and distances
            ind_dist = list(map(ind_dict, neighbours.values, neighbours_dist.values))
            
            
        # b) Local radius method
        if method == 'radius':
            print("Identifying neighbours within " + str(radius) + " pixels of every cell")
            kdt = BallTree(data, metric='euclidean')
            ind, dist = kdt.query_radius(data, r=radius, return_distance= True)
            # Remove self from ind
            for i in range(0, len(ind)): ind[i] = np.delete(ind[i], np.argwhere(ind[i] == i))#remove self
            for i in range(0, len(ind)): dist[i] =  dist[i][dist[i] != 0]#remove self distance
            # Merge the indeces and distance into a single dictionary
            ind_dict = lambda x,y: dict(zip(x, y)) # Function to create a dict of indeces and distances
            ind_dist = list(map(ind_dict, ind.tolist(),dist.tolist()))
        
        # Modify the dictionary to custom range for normalization 
        def range_normalize (data, upper_limit, lower_limit): # Function to normalize based on custom range
            
            if len(data) >= 2:
                # calculate data range
                x = (max (data.values()) - min (data.values())) / len (data)
                # calculate normalization factor range
                y = (upper_limit - lower_limit) / len (data)
                # Step-2: For each data point calculate the normalization factor
                xij = []
                for i in data:
                    xij = np.append(xij, ((max (data.values()) - data[i]) * y) / x)
                # Step-3: Compute the normalized values from the factors determined
                yij = []
                for j in xij:
                    yij = xij + lower_limit
                # modify the data object to reflect the new normlized values
                modified_data = dict(zip(data.keys(), yij))
            elif len(data) == 1:
                data[list(data.keys())[0]] = upper_limit
                modified_data = data
            elif data == {}:
                modified_data = {}
            
            return modified_data
        
        # Define range
        n = lambda x: range_normalize(x, 1, 0)
        # Apply function
        normalized_dist = list(map(n, ind_dist))
        
        # Normalize the spatial weights based on total number of neighbours per cell
        def norm (data):
            norm_data = {k: v/sum(data.values()) for k, v in data.items()}
            return norm_data
        
        # Run the function on all cell neighbourhoods  
        cell_number_normalized_dist = list(map(norm, normalized_dist))
        
        ## Spatial Lag calculation
        # Convert the dictionary to a martix
        idx = range(len(cell_number_normalized_dist))
        wn_matrix = pd.DataFrame(cell_number_normalized_dist, columns=idx, index=idx).fillna(0)
        wn_matrix.columns = data.index # add the cell name
        wn_matrix.index = data.index  # add the cell name
        
        # Convert the matrix to a sparce matrix
        wn_matrix_sparse = csr_matrix(wn_matrix)
        
        # Calculation of spatial lag
        if use_raw==True:
            spatial_lag = pd.DataFrame(wn_matrix_sparse * np.log1p(adata_subset.raw.X), columns = adata_subset.var.index, index=adata_subset.obs.index)
        else:
            spatial_lag = pd.DataFrame(wn_matrix_sparse * adata_subset.X, columns = adata_subset.var.index, index=adata_subset.obs.index)
        
        # return value
        return spatial_lag
    
    # Subset a particular image if needed
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
    
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_expression_internal = lambda x: spatial_expression_internal(adata_subset=x, 
                                                                x_coordinate=x_coordinate, 
                                                                y_coordinate=y_coordinate, 
                                                                method=method, radius=radius, 
                                                                knn=knn, imageid=imageid, 
                                                                use_raw=use_raw, subset=subset,
                                                                label=label) 
    all_data = list(map(r_spatial_expression_internal, adata_list)) # Apply function 
    
    # Merge all the results into a single dataframe    
    result = []
    for i in range(len(all_data)):
        result.append(all_data[i])
    result = pd.concat(result, join='outer')  
    
    # Reindex the cells
    result = result.fillna(0)
    result = result.reindex(adata.obs.index)
    
    # Add to adata
    adata.uns[label] = result
    
    # Return        
    return adata




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 13:07:38 2020
@author: Ajit Johnson Nirmal
visualization option for understanding average shortest distance between phenotypes of interest.
Pre-run `sm.tl.spatial_distance` before running this function.
"""

# library
import pandas as pd
import matplotlib
import numpy as np
import seaborn as sns; sns.set(color_codes=True)
sns.set_style("white")


def spatial_distance (adata, spatial_distance='spatial_distance',phenotype='phenotype',imageid='imageid',log=False,
                      method='heatmap',heatmap_summarize=True,heatmap_na_color='grey',heatmap_cmap='vlag_r',
                      heatmap_row_cluster=False,heatmap_col_cluster=False,heatmap_standard_scale=0,
                      distance_from=None,distance_to=None,x_axis = None,y_axis = None,facet_by = None,plot_type = None,
                      **kwargs):
    """
    

    Parameters
    ----------
    adata : AnnData object

    spatial_distance : string, optional
        In order to locate the spatial_distance data within the AnnData object please provide the output 
        label/columnname of `sm.tl.spatial_distance` function. The default is 'spatial_distance'.
    phenotype : string, required
        Column name of the column containing the phenotype information. 
        It could also be any categorical assignment given to single cells. The default is 'phenotype'.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    log : bool, optional
        Convert distance to log scale. The default is False.
    method : string, optional
        Three options are available.
        1) heatmap - generates a heatmap of average shortest distance between all phenotypes.
        2) numeric - can be used to generate boxplot, violin plot etc between a given set of phenotypes.
        3) distribution - can be used to generate distribution plots between a given set of phenotypes.
        The default is 'heatmap'.
    heatmap_summarize : bool, optional
        In the event multiple images are present in the dataset, True allows to calculate the 
        average across all the images. The default is True.
    heatmap_na_color : string, optional
        Color for NA values within the heatmap. The default is 'grey'.
    heatmap_cmap : string, optional
        Color map to use for continous variables. 
        Can be a name or a Colormap instance (e.g. 'magma', 'viridis'). The default is 'vlag_r'.
    heatmap_row_cluster : bool, optional
        Cluster Rows. The default is False.
    heatmap_col_cluster : bool, optional
        Cluster Columns. The default is False.
    heatmap_standard_scale : int, optional
        Either 0 (rows) or 1 (columns). Whether or not to standardize that dimension, 
        meaning for each row or column, subtract the minimum and divide each by its maximum. The default is 0.
    distance_from : string, optional
        In the event of using method = 'numeric' or 'distribution', this argument is required.
        Pass a phenotype of interest. If distance_from is provided and distance_to is not provided,
        the function will plot the average distance from the phenotype of interest to all
        phenotypes present within the dataset. The default is None.
    distance_to : string, optional
        In the event of using method = 'numeric' or 'distribution', this argument is required.
        Pass a phenotype of interest. The function will plot the average shortest between two phenotypes of
        interest (distance_from and distance_to). The default is None.
    x_axis : string, optional
        In the event of using method = 'numeric' or 'distribution', this argument is required.
        This determines the elements present in the x-axis of the resultant plot. 
        Allowed arguments are: 'group', 'distance', 'imageid'.
        The default is `distance`.
    y_axis : string, optional
        In the event of using method = 'numeric' or 'distribution', this argument is required.
        This determines the elements present in the y-axis of the numeric plot and if the user uses the distribution
        plot this argument is used to overlaying multiple categories within the same distribution plot.
        Allowed arguments are: 'group', 'distance', 'imageid'.
        The default is `group`.
    facet_by : string, optional
         In the event of using method = 'numeric' or 'distribution', this argument can be used to
         generate sub-plots. Allowed arguments are: 'group', 'imageid'.
         The default is None.
    plot_type : string, optional
        In the event of using method = 'numeric' or 'distribution', this argument is required.
        For `numeric` plot, the following options are available: “strip”, “swarm”, “box”, “violin”, “boxen”, “point”, “bar”, or “count”.
        For `distribution` plot, the following options are available: “hist”, “kde”, “ecdf”.
        The default for `numeric` plot is 'boxen'.
        The default for `distribution` plot is 'kde`.
    **kwargs : dict
        Are passed to sns.clustermap. Pass other parameters that works with `sns.clustermap`, `sns.catplot` or `sns.displot`
        e.g. `linecolor='black'`.

    Returns
    -------
    Heatmap or Numeric Plot or Distribution Plot.
    
    Example
    -------
    # summary heatmap
    sm.pl.spatial_distance (adata)
    
    # Heatmap without summarizing the individual images
    sm.pl.spatial_distance (adata, heatmap_summarize=False, imageid='ImageId')
    
    # Numeric plot of shortest distance of phenotypes from tumor cells
    sm.pl.spatial_distance (adata, method='numeric',distance_from='Tumor CD30+',imageid='ImageId')
    
    # Distribution plot of shortest distance of phenotypes from tumor cells
    sm.pl.spatial_distance (adata, method='distribution',distance_from='Tumor CD30+',imageid='ImageId', x_axis="distance", y_axis="imageid", plot_type="kde")
    
    # Numeric plot of shortest distance of phenotypes from tumor cells to M2 Macrophages
    sm.pl.spatial_distance (adata, method='numeric',distance_from='Tumor CD30+',distance_to = 'M2 Macrophages', imageid='ImageId')
    
    # Distribution plot of shortest distance of phenotypes from tumor cells to M2 Macrophages
    sm.pl.spatial_distance (adata, method='distribution',distance_from='Tumor CD30+',distance_to = 'M2 Macrophages',imageid='ImageId')


    """
    
            
    # set color for heatmap
    cmap_updated = matplotlib.cm.get_cmap(heatmap_cmap)
    cmap_updated.set_bad(color=heatmap_na_color)
    
    
    # Copy the spatial_distance results from anndata object
    try:
        diatance_map = adata.uns[spatial_distance].copy()
    except KeyError:
        raise ValueError('spatial_distance not found- Please run sm.tl.spatial_distance first')
    
    # Convert distance to log scale if user requests
    if log is True:
        diatance_map = np.log1p(diatance_map)
            
    # Method
    if method=='heatmap':
        if heatmap_summarize is True:
            # create the necessary data
            data = pd.DataFrame({'phenotype': adata.obs[phenotype]})
            data = pd.merge(data, diatance_map, how='outer',left_index=True, right_index=True) # merge with the distance map
            k = data.groupby(['phenotype']).mean() # collapse the whole dataset into mean expression
            k = k[k.index]
        else:
            # create new naming scheme for the phenotypes
            non_summary = pd.DataFrame({'imageid': adata.obs[imageid], 'phenotype': adata.obs[phenotype]})
            non_summary['imageid'] = non_summary['imageid'].astype(str) # convert the column to string
            non_summary['phenotype'] = non_summary['phenotype'].astype(str) # convert the column to string
            non_summary['image_phenotype'] = non_summary['imageid'].str.cat(non_summary['phenotype'],sep="_")
            # Merge distance map with phenotype
            data = pd.DataFrame(non_summary[['image_phenotype']])
            data = pd.merge(data, diatance_map, how='outer',left_index=True, right_index=True)
            k = data.groupby(['image_phenotype']).mean()
            k = k.sort_index(axis=1)
        # Generate the heatmap
        mask = k.isnull() # identify the NAN's for masking 
        k = k.fillna(0) # replace nan's with 0 so that clustering will work
        # Heatmap
        sns.clustermap(k, cmap=heatmap_cmap, row_cluster=heatmap_row_cluster, 
                       col_cluster=heatmap_col_cluster, mask=mask, 
                       standard_scale=heatmap_standard_scale, **kwargs)
    else:
            
        # condition-1
        if distance_from is None and distance_to is None:
            raise ValueError('Please include distance_from and/or distance_to parameters to use this method')
                     
        # condition-2
        if distance_from is None and distance_to is not None:
            raise ValueError('Please `distance_from` parameters to use this method')
            
        # condition-3
        if distance_to is not None:
            # convert input to list if needed
            if isinstance(distance_to, str):
                distance_to = [distance_to]
        
        # Start
        pheno_df = pd.DataFrame({'imageid': adata.obs[imageid], 'phenotype': adata.obs[phenotype]}) #image id and phenotype
        data = pd.merge(pheno_df, diatance_map, how='outer',left_index=True, right_index=True) # merge with the distance map
        data = data[data['phenotype'] == distance_from] # subset the pheno of interest
        
        if distance_to is not None:
            data = data[distance_to] # drop columns that are not requested in distance_to
        else:
            data = data.drop(['phenotype','imageid'], axis=1) # drop the phenotype column before stacking
         
        d = data.stack().reset_index() # collapse everything to one column
        d.columns = ['cellid', 'group', 'distance']
        d = pd.merge(d, pheno_df, left_on='cellid', right_index=True) # bring back the imageid and phenotype
         
        # Convert columns to str
        for col in ['imageid', 'group','phenotype']:
            d[col] = d[col].astype(str)
                
        # Convert columns to categorical so that it drops unused categories
        for col in ['imageid', 'group','phenotype']:
            d[col] = d[col].astype('category')
        
        
        # Plotting
        if method=='numeric':
            if x_axis is None and y_axis is None and facet_by is None and plot_type is None:
                sns.catplot(data=d, x="distance", y="group", col="imageid", kind="boxen", **kwargs)
            else:
                sns.catplot(data=d, x=x_axis, y=y_axis, col=facet_by, kind=plot_type, **kwargs)
            
        if method=='distribution':
            if x_axis is None and y_axis is None and facet_by is None and plot_type is None:
                sns.displot(data=d, x="distance", hue="imageid",  col="group", kind="kde", **kwargs)
            else:
                sns.displot(data=d, x=x_axis, hue=y_axis, col=facet_by, kind=plot_type,**kwargs)    
            



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19 20:03:01 2020
@author: Ajit Johnson Nirmal
Function that computes how likely a celltype is found close to another 
"""

# Import library
import pandas as pd
from sklearn.neighbors import BallTree
import numpy as np
from joblib import Parallel, delayed
import scipy
from functools import reduce


# Function
def spatial_interaction (adata,x_coordinate='X_centroid',y_coordinate='Y_centroid',
                         phenotype='phenotype',
                         method='radius', radius=30, knn=10,
                         permutation=1000,
                         imageid='imageid',subset=None,
                         pval_method='histocat',
                         label='spatial_interaction'):
    """
    

    Parameters
    ----------
    adata : AnnData object

    x_coordinate : float, required
        Column name containing the x-coordinates values. The default is 'X_centroid'.
    y_coordinate : float, required
        Column name containing the y-coordinates values. The default is 'Y_centroid'.
    phenotype : string, required
        Column name of the column containing the phenotype information. 
        It could also be any categorical assignment given to single cells. The default is 'phenotype'.
    method : string, optional
        Two options are available: a) 'radius', b) 'knn'.
        a) radius - Identifies the neighbours within a given radius for every cell.
        b) knn - Identifies the K nearest neigbours for every cell.
        The default is 'radius'.
    radius : int, optional
        The radius used to define a local neighbhourhood. The default is 30.
    knn : int, optional
        Number of cells considered for defining the local neighbhourhood. The default is 10.
    permutation : int, optional
        The number of permutations to be performed for calculating the P-Value. The default is 1000.
    imageid : string, optional
        Column name of the column containing the image id. The default is 'imageid'.
    subset : string, optional
        imageid of a single image to be subsetted for analyis. The default is None.
    pval_method : string, optional
        Two options are available: a) 'histocat', b) 'zscore'.
        a) P-values are calculated by subtracting the permuted mean from the observed mean 
        divided by the number of permutations as described in the histoCAT manuscript (Denis et.al, Nature Methods 2017)
        b) zscores are calculated from the mean and standard deviation and further p-values are
        derived by fitting the observed values to a normal distribution. The default is 'histocat'.
    label : string, optional
        Key for the returned data, stored in `adata.obs`. The default is 'spatial_interaction'.

    Returns
    -------
    adata : AnnData object
        Updated AnnData object with the results stored in `adata.obs['spatial_aggregate']`.
    
    Example
    -------
    # Using the radius method to identify local neighbours and histocat to compute P-values
    adata = sm.tl.spatial_interaction(adata, method='radius', radius=30,pval_method='histocat',
                                      imageid='ImageId',x_coordinate='X_position',y_coordinate='Y_position')
    
    # Using the KNN method to identify local neighbours and zscore to compute P-values
    adata = sm.tl.spatial_interaction(adata, method='knn', radius=30,pval_method='zscore',
                                      imageid='ImageId',x_coordinate='X_position',y_coordinate='Y_position')

    """
    
    
    def spatial_interaction_internal (adata_subset,x_coordinate,y_coordinate,
                                      phenotype,
                                      method, radius, knn,
                                      permutation, 
                                      imageid,subset,
                                      pval_method):
        
        print("Processing Image: " + str(adata_subset.obs[imageid].unique()))
        
        # Create a dataFrame with the necessary inforamtion
        data = pd.DataFrame({'x': adata_subset.obs[x_coordinate], 'y': adata_subset.obs[y_coordinate], 'phenotype': adata_subset.obs[phenotype]})
        
        
        # Identify neighbourhoods based on the method used
        # a) KNN method
        if method == 'knn':
            print("Identifying the " + str(knn) + " nearest neighbours for every cell")
            tree = BallTree(data[['x','y']], leaf_size= 2)
            ind = tree.query(data[['x','y']], k=knn, return_distance= False)
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            neighbours.drop(0, axis=1, inplace=True) # Remove self neighbour
            
        # b) Local radius method
        if method == 'radius':
            print("Identifying neighbours within " + str(radius) + " pixels of every cell")
            kdt = BallTree(data[['x','y']], metric='euclidean') 
            ind = kdt.query_radius(data[['x','y']], r=radius, return_distance=False)
            for i in range(0, len(ind)): ind[i] = np.delete(ind[i], np.argwhere(ind[i] == i))#remove self
            neighbours = pd.DataFrame(ind.tolist(), index = data.index) # neighbour DF
            
        # Map Phenotypes to Neighbours
        # Loop through (all functionized methods were very slow)
        phenomap = dict(zip(list(range(len(ind))), data['phenotype'])) # Used for mapping
        print("Mapping phenotype to neighbors")
        for i in neighbours.columns:
            neighbours[i] = neighbours[i].dropna().map(phenomap, na_action='ignore')
            
        # Drop NA
        neighbours = neighbours.dropna(how='all')
        
        # Collapse all the neighbours into a single column
        n = pd.DataFrame(neighbours.stack(), columns = ["neighbour_phenotype"])
        n.index = n.index.get_level_values(0) # Drop the multi index
        
        # Merge with real phenotype
        n = n.merge(data['phenotype'], how='inner', left_index=True, right_index=True)
        
        # Permutation
        print('Performing '+ str(permutation) + ' permutations')
    
        def permutation_pval (data):
            data = data.assign(neighbour_phenotype=np.random.permutation(data['neighbour_phenotype']))
            #data['neighbour_phenotype'] = np.random.permutation(data['neighbour_phenotype'])
            data_freq = data.groupby(['phenotype','neighbour_phenotype']).size().unstack()
            data_freq = data_freq.fillna(0).stack().values 
            return data_freq
        
        # Apply function
        final_scores = Parallel(n_jobs=-1)(delayed(permutation_pval)(data=n) for i in range(permutation)) 
        perm = pd.DataFrame(final_scores).T
        
        # Consolidate the permutation results
        print('Consolidating the permutation results')
        # Calculate P value
        # real
        n_freq = n.groupby(['phenotype','neighbour_phenotype']).size().unstack().fillna(0).stack() 
        # permutation
        mean = perm.mean(axis=1)
        std = perm.std(axis=1)
        # P-value calculation
        if pval_method == 'histocat':
            # real value - prem value / no of perm 
            p_values = abs(n_freq.values - mean) / (permutation+1)
            p_values = p_values[~np.isnan(p_values)].values
        if pval_method == 'zscore':
            z_scores = (n_freq.values - mean) / std        
            z_scores[np.isnan(z_scores)] = 0
            p_values = scipy.stats.norm.sf(abs(z_scores))*2
            p_values = p_values[~np.isnan(p_values)]
            
        # Compute Direction of interaction (interaction or avoidance)
        direction = ((n_freq.values - mean) / abs(n_freq.values - mean)).fillna(1)

        # Normalize based on total cell count
        k = n.groupby(['phenotype','neighbour_phenotype']).size().unstack().fillna(0)
        total_cell_count = data['phenotype'].value_counts()
        total_cell_count = total_cell_count.reindex(k.columns).values
        k_max = k.div(total_cell_count, axis = 0)
        k_max = k_max.div(k_max.max(axis=1), axis=0).stack()
        
        # DataFrame with the neighbour frequency and P values
        count = (k_max.values * direction).values # adding directionallity to interaction
        neighbours = pd.DataFrame({'count': count,'p_val': p_values}, index = k_max.index)
        #neighbours.loc[neighbours[neighbours['p_val'] > p_val].index,'count'] = np.NaN
        #del neighbours['p_val']
        neighbours.columns = [adata_subset.obs[imageid].unique()[0], 'pvalue_' + str(adata_subset.obs[imageid].unique()[0])]
        neighbours = neighbours.reset_index()
        #neighbours = neighbours['count'].unstack()
        
        # return
        return neighbours
          
      
    # subset a particular subset of cells if the user wants else break the adata into list of anndata objects
    if subset is not None:
        adata_list = [adata[adata.obs[imageid] == subset]]
    else:
        adata_list = [adata[adata.obs[imageid] == i] for i in adata.obs[imageid].unique()]
    
    
    # Apply function to all images and create a master dataframe
    # Create lamda function 
    r_spatial_interaction_internal = lambda x: spatial_interaction_internal (adata_subset=x, x_coordinate=x_coordinate, y_coordinate=y_coordinate, phenotype=phenotype, method=method,  radius=radius, knn=knn, permutation=permutation, imageid=imageid,subset=subset,pval_method=pval_method) 
    all_data = list(map(r_spatial_interaction_internal, adata_list)) # Apply function 
    

    # Merge all the results into a single dataframe    
    df_merged = reduce(lambda  left,right: pd.merge(left,right,on=['phenotype', 'neighbour_phenotype'], how='outer'), all_data)
    

    # Add to anndata
    adata.uns[label] = df_merged
    
    # return
    return adata








from ._gmm_gate import gmm_gate
from ._gate_finder import gate_finder
from ._image_viewer import image_viewer
from ._spatial_interaction import spatial_interaction
from ._spatial_distance import spatial_distance


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar  2 19:56:08 2020
@author: Ajit Johnson Nirmal
Cell Phenotyping
"""

# Library
import numpy as np
import pandas as pd


def phenotype_cells (adata, phenotype, gate = 0.5, label="phenotype", imageid='imageid',
                      pheno_threshold_percent=None, pheno_threshold_abs=None):
    """


    Parameters
    ----------
    adata : AnnData Object
    phenotype : DataFrame
        A gating strategy for phenotyping the cells. Example provided here <add link>.
    gate : int, optional (The default is 0.5)
        By default rescale function, scales the data such that values above 0.5 are considered positive cells.
    label : string, optional (The default is "phenotype")
        Name the column underwhich the final phenotype calling will be saved.
    imageid : string, optional (The default is 'imageid')
        Name of the column that contains the unique imageid. This is only utilized
        when the user uses `pheno_threshold_percent` or `pheno_threshold_abs` parameters.
    pheno_threshold_percent : float, optional (The default is None)
        Accepts values between (0-100). If any particular phenotype is below the user defined threshold,
        it is recategorised as 'unknown. Generally used to deal with low background false positives.
    pheno_threshold_abs : int, optional (The default is None)
        Serves the same purpose as that of pheno_threshold_percent. However, an absolute
        number can be passed. For example, if user passes in 10- any phenotype that contains
        less than 10 cells will be recategorized as unknown.

    Returns
    -------
    AnnData
        Updated AnnData object with the phenotype calls for each cell.

    Example
    -------
    phenotype = pd.read_csv('path/to/csv/file/')
    adata = sm.tl.phenotype_cells (adata, phenotype=phenotype, gate = 0.5, label="phenotype")

    """

    # Create a dataframe from the adata object
    data = pd.DataFrame(adata.X, columns = adata.var.index, index= adata.obs.index)

    # Function to calculate the phenotype scores
    def phenotype_cells (data,phenotype,gate,group):

        # Subset the phenotype based on the group
        phenotype = phenotype[phenotype.iloc[:,0] == group]

        # Parser to parse the CSV file into four categories
        def phenotype_parser (p, cell):
            # Get the index and subset the phenotype row being passed in
            location = p.iloc[:,1] == cell
            idx = [i for i, x in enumerate(location) if x][0]
            phenotype = p.iloc[idx,:]
            # Calculate
            pos = phenotype[phenotype == 'pos'].index.tolist()
            neg = phenotype[phenotype == 'neg'].index.tolist()
            anypos = phenotype[phenotype == 'anypos'].index.tolist()
            anyneg = phenotype[phenotype == 'anyneg'].index.tolist()
            allpos = phenotype[phenotype == 'allpos'].index.tolist()
            allneg = phenotype[phenotype == 'allneg'].index.tolist()
            return {'pos': pos, 'neg': neg ,'anypos': anypos, 'anyneg': anyneg, 'allpos': allpos, 'allneg': allneg}
            #return pos, neg, anypos, anyneg

        # Run the phenotype_parser function on all rows
        p_list = phenotype.iloc[:,1].tolist()
        r_phenotype = lambda x: phenotype_parser(cell=x, p=phenotype) # Create lamda function
        all_phenotype = list(map(r_phenotype, p_list)) # Apply function
        all_phenotype = dict(zip(p_list, all_phenotype)) # Name the lists

        # Define function to check if there is any marker that does not satisfy the gate
        def gate_satisfation_lessthan (marker, data, gate):
            fail = np.where(data[marker] < gate, 1, 0) # 1 is fail
            return fail
        # Corresponding lamda function
        r_gate_satisfation_lessthan = lambda x: gate_satisfation_lessthan(marker=x, data=data, gate=gate)

        # Define function to check if there is any marker that does not satisfy the gate
        def gate_satisfation_morethan (marker, data, gate):
            fail = np.where(data[marker] > gate, 1, 0)
            return fail
        # Corresponding lamda function
        r_gate_satisfation_morethan = lambda x: gate_satisfation_morethan(marker=x, data=data, gate=gate)

        def prob_mapper (data, all_phenotype, cell, gate):

            print("Phenotyping " + str(cell))

            # Get the appropriate dict from all_phenotype
            p = all_phenotype[cell]

            # Identiy the marker used in each category
            pos = p.get('pos')
            neg = p.get('neg')
            anypos = p.get('anypos')
            anyneg = p.get('anyneg')
            allpos = p.get('allpos')
            allneg = p.get('allneg')

            # Perform computation for each group independently
            # Positive marker score
            if len(pos) != 0:
                pos_score = data[pos].mean(axis=1).values
                pos_fail = list(map(r_gate_satisfation_lessthan, pos)) if len(pos) > 1 else []
                pos_fail = np.amax(pos_fail, axis=0) if len(pos) > 1 else []
            else:
                pos_score = np.repeat(0, len(data))
                pos_fail = []

            # Negative marker score
            if len(neg) != 0:
                neg_score = (1-data[neg]).mean(axis=1).values
                neg_fail = list(map(r_gate_satisfation_morethan, neg)) if len(neg) > 1 else []
                neg_fail = np.amax(neg_fail, axis=0) if len(neg) > 1 else []
            else:
                neg_score = np.repeat(0, len(data))
                neg_fail = []

            # Any positive score
            anypos_score = np.repeat(0, len(data)) if len(anypos) == 0 else data[anypos].max(axis=1).values

            # Any negative score
            anyneg_score = np.repeat(0, len(data)) if len(anyneg) == 0 else (1-data[anyneg]).max(axis=1).values

            # All positive score
            if len(allpos) != 0:
                allpos_score = data[allpos]
                allpos_score['score'] = allpos_score.max(axis=1)
                allpos_score.loc[(allpos_score < gate).any(axis = 1), 'score'] = 0
                allpos_score = allpos_score['score'].values + 0.01 # A small value is added to give an edge over the matching positive cell
            else:
                allpos_score = np.repeat(0, len(data))


            # All negative score
            if len(allneg) != 0:
                allneg_score = 1- data[allneg]
                allneg_score['score'] = allneg_score.max(axis=1)
                allneg_score.loc[(allneg_score < gate).any(axis = 1), 'score'] = 0
                allneg_score = allneg_score['score'].values + 0.01
            else:
                allneg_score = np.repeat(0, len(data))


            # Total score calculation
            # Account for differences in the number of categories used for calculation of the final score
            number_of_non_empty_features = np.sum([len(pos) != 0,
                                                len(neg) != 0,
                                                len(anypos) != 0,
                                                len(anyneg) != 0,
                                                len(allpos) != 0,
                                                len(allneg) != 0])

            total_score = (pos_score + neg_score + anypos_score + anyneg_score + allpos_score + allneg_score) / number_of_non_empty_features

            return {cell: total_score, 'pos_fail': pos_fail ,'neg_fail': neg_fail}
            #return total_score, pos_fail, neg_fail


        # Apply the fuction to get the total score for all cell types
        r_prob_mapper = lambda x: prob_mapper (data=data, all_phenotype=all_phenotype, cell=x, gate=gate) # Create lamda function
        final_scores = list(map(r_prob_mapper, [*all_phenotype])) # Apply function
        final_scores = dict(zip([*all_phenotype], final_scores)) # Name the lists

        # Combine the final score to annotate the cells with a label
        final_score_df = pd.DataFrame()
        for i in [*final_scores]:
            df = pd.DataFrame(final_scores[i][i])
            final_score_df= pd.concat([final_score_df, df], axis=1)
        # Name the columns
        final_score_df.columns = [*final_scores]
        final_score_df.index = data.index
        # Add a column called unknown if all markers have a value less than the gate (0.5)
        unknown = group + str('-rest')
        final_score_df[unknown] = (final_score_df < gate).all(axis=1).astype(int)

        # Name each cell
        labels = final_score_df.idxmax(axis=1)

        # Group all failed instances (i.e. when multiple markers were given
        # any one of the marker fell into neg or pos zones of the gate)
        pos_fail_all = pd.DataFrame()
        for i in [*final_scores]:
            df = pd.DataFrame(final_scores[i]['pos_fail'])
            df.columns = [i] if len(df) != 0 else []
            pos_fail_all= pd.concat([pos_fail_all, df], axis=1)
        pos_fail_all.index = data.index if len(pos_fail_all) != 0 else []
        # Same for Neg
        neg_fail_all = pd.DataFrame()
        for i in [*final_scores]:
            df = pd.DataFrame(final_scores[i]['neg_fail'])
            df.columns = [i] if len(df) != 0 else []
            neg_fail_all= pd.concat([neg_fail_all, df], axis=1)
        neg_fail_all.index = data.index if len(neg_fail_all) != 0 else []


        # Modify the labels with the failed annotations
        if len(pos_fail_all) != 0:
            for i in pos_fail_all.columns:
                labels[(labels == i) & (pos_fail_all[i] == 1)] = 'likely-' + i
        # Do the same for negative
        if len(neg_fail_all) != 0:
            for i in neg_fail_all.columns:
                labels[(labels == i) & (neg_fail_all[i] == 1)] = 'likely-' + i

        # Retun the labels
        return labels

    # Create an empty dataframe to hold the labeles from each group
    phenotype_labels = pd.DataFrame()

    # Loop through the groups to apply the phenotype_cells function
    for i in phenotype.iloc[:,0].unique():

        if phenotype_labels.empty:
            phenotype_labels = pd.DataFrame(phenotype_cells(data = data, group = i, phenotype=phenotype, gate=gate))
            phenotype_labels.columns = [i]

        else:
            # Find the column with the cell-type of interest
            column_of_interest = [] # Empty list to hold the column name
            try:
                column_of_interest = phenotype_labels.columns[phenotype_labels.eq(i).any()]
            except:
                pass
            # If the cell-type of interest was not found just add NA
            if len(column_of_interest) == 0:
                phenotype_labels[i] = np.nan
            else:
                #cells_of_interest = phenotype_labels[phenotype_labels[column_of_interest] == i].index
                cells_of_interest = phenotype_labels[phenotype_labels[column_of_interest].eq(i).any(axis=1)].index
                d = data.loc[cells_of_interest]
                print("-- Subsetting " + str(i))
                phenotype_l = pd.DataFrame(phenotype_cells(data = d, group = i, phenotype=phenotype, gate=gate), columns = [i])
                phenotype_labels = phenotype_labels.merge(phenotype_l, how='outer', left_index=True, right_index=True)

    # Rearrange the rows back to original
    phenotype_labels = phenotype_labels.reindex(data.index)
    phenotype_labels = phenotype_labels.replace('-rest', np.nan, regex=True)

    print("Consolidating the phenotypes across all groups")
    phenotype_labels_Consolidated = phenotype_labels.fillna(method='ffill', axis = 1)
    phenotype_labels[label] = phenotype_labels_Consolidated.iloc[:,-1].values

    # replace nan to 'other cells'
    phenotype_labels[label] = phenotype_labels[label].fillna('Unknown')

    # Apply the phenotype threshold if given
    if pheno_threshold_percent or pheno_threshold_abs is not None:
        p = pd.DataFrame(phenotype_labels[label])
        q = pd.DataFrame(adata.obs[imageid])
        p = q.merge(p, how='outer', left_index=True, right_index=True)

        # Function to remove phenotypes that are less than the given threshold
        def remove_phenotype(p, ID, pheno_threshold_percent, pheno_threshold_abs):
            d = p[p[imageid] == ID]
            x = pd.DataFrame(d.groupby([label]).size())
            x.columns = ['val']
            # FInd the phenotypes that are less than the given threshold
            if pheno_threshold_percent is not None:
                fail = list(x.loc[x['val'] < x['val'].sum() * pheno_threshold_percent/100].index)
            if pheno_threshold_abs is not None:
                fail = list(x.loc[x['val'] < pheno_threshold_abs].index)
            d[label] = d[label].replace(dict(zip(fail, np.repeat('Unknown',len(fail)))))
            # Return
            return d

        # Apply function to all images
        r_remove_phenotype = lambda x: remove_phenotype (p=p, ID=x,
                                                         pheno_threshold_percent=pheno_threshold_percent,
                                                         pheno_threshold_abs=pheno_threshold_abs) # Create lamda function
        final_phrnotypes= list(map(r_remove_phenotype, list(p[imageid].unique()))) # Apply function

        final_phrnotypes = pd.concat(final_phrnotypes, join='outer')
        phenotype_labels = final_phrnotypes.reindex(adata.obs.index)


    # Return to adata
    adata.obs[label] = phenotype_labels[label]

    #for i in phenotype_labels.columns:
    #    adata.obs[i] = phenotype_labels[i]

    return adata



# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 21:00:57 2020
@author: Ajit Johnson Nirmal
Helper function to save andata object as a csv file
"""

# Lib
import pandas as pd
import numpy as np


# Function

def scimap_to_csv (adata, data_type='raw'):
    """
    

    Parameters
    ----------
    adata : AnnData Object

    data_type : string, optional
        Three options are available:
        1) 'raw' - The raw data will be returned.
        2) 'log' - The raw data converted to log scale using `np.log1p` will be returned.
        3) 'scaled' - If you have scaled the data using the `sm.pp.rescale`, that will be
        returned. Please note, if you have not scaled the data, whatever is within
        `adata.X` will be returned.
        The default is 'raw'.

    Returns
    -------
    merged : DataFrame
        A single dataframe containing the expression and metadata will be returned.
        
    Example
    -------
    data = sm.hl.scimap_to_csv (adata, data_type='raw')

    """
    
    # Expression matrix
    if data_type is 'raw':
        data = pd.DataFrame(adata.raw.X, index=adata.obs.index, columns=adata.var.index)
    if data_type is 'log':
        data = pd.DataFrame(np.log1p(adata.raw.X), index=adata.obs.index, columns=adata.var.index)
    if data_type is 'scaled':
        data = pd.DataFrame(adata.X, index=adata.obs.index, columns=adata.var.index)
    
    # Metadata
    meta = pd.DataFrame(adata.obs)
    
    # Merge the two dataframes
    merged = pd.concat([data, meta], axis=1, sort=False)
    
    # Add a column to save cell-id
    merged['CellID'] = merged.index
    
    # reset index
    merged = merged.reset_index(drop=True)
    
    # return
    return merged




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Mar 22 13:08:26 2020
@author: Ajit Johnson Nirmal
Rename any string within a column to anything
"""

import functools 
import re


def rename (adata, rename, from_column='phenotype', to_column='phenotype_renamed'):
    """
    
    Parameters
    ----------
    adata : AnnData object
    rename : dict
        Pass a dictionary with 'values' as elements that need to be altered and 'keys' as the elements that they need to be transformed into.
    from_column : string, required
        Column that need to be modified. The default is 'phenotype'.
    to_column : string, required
        Modified names will be stored in a new column with this name. The default is 'phenotype_renamed'.

    Returns
    -------
    adata : Modified AnnData Object
        DESCRIPTION.
    
    Example
    -------
    rename= {'tumor': ['cd45 neg tumor', 'cd8 tumor', 'cd4 tumor'],
             'macrophages': ['m1 macrophages', 'm2 macrophages']}
    Here we are renaming cd45 neg tumor, cd8 tumor and cd4 tumor into 'tumor' and 
    m1 macrophages and m2 macrophages into macrophages
    
    adata = sm.hl.rename_clusters (adata, rename, from_column='phenotype', to_column='phenotype_renamed')

    """
    
    # Get the from_column
    rename_from = list(adata.obs[from_column].values)
    
    # Split multiple renaming events into independent events
    name = functools.reduce( lambda x,y: dict(x, **y), (dict(map(lambda x: (x,i), rename[i])) for i in rename))
     
    # Rename
    for i in name:
        print ('Renaming ' + str(i) + ' to ' + str(name[i]))
        #rename_from = [x.replace(i, name[i]) for x in rename_from]
        s = str(i)
        s = s.replace('+', '\+')
        #rename_from = [re.sub(r'^\b%s\b$' % s,  name[i], j) for j in rename_from]
        rename_from = [re.sub(r'^\b%s$' % s,  name[i], j) for j in rename_from]
        
    
    # Append to adata as a new column
    adata.obs[to_column] = rename_from
    
    # Return
    return adata



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 10 09:32:49 2020
@author: Ajit Johnson Nirmal
Tests
"""

import pytest
import sys, os

#os.chdir ("/Users/aj/Dropbox (Partners HealthCare)/packages/scimap")

@pytest.fixture
def adata():
    from scimap.preprocessing._mcmicro_to_scimap import mcmicro_to_scimap
    image_path = [os. getcwd() + '/scimap/tests/_data/example_data.csv']
    adata = mcmicro_to_scimap (image_path, split='X_position', drop_markers=['BG1', 'BG2', 'BG3'])
    return adata


def test_mcmicro_to_scimap(adata):
    assert adata.shape == (3029, 33)

def test_rescale(adata):
    import pandas as pd
    import numpy as np
    from scimap.preprocessing._rescale import rescale

    # test rescaling data
    manual_gate = pd.DataFrame({'marker': ['CD3D', 'KI67'], 'gate': [7, 8]})
    adata = rescale (adata, gate=manual_gate, failed_markers=['CD20', 'CD21'])
    a = np.round(adata[:,'CD3D'].X[0],2)

    # test
    assert a == 0.37



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 26 12:04:17 2020
@author: Ajit Johnson Nirmal
Helper function to annotate cells based on positivity/negativity of defined markers. Users can classify
the entire data or a subset of data that has been previously phenotyped/clustered/
"""

# Library
import pandas as pd
import numpy as np


# Functions
def classify (adata, pos=None, neg=None, classify_label='passed_classify', 
              phenotype='phenotype', subclassify_phenotype=None, 
              collapse_failed=True, label="classify"):
    """
    

    Parameters
    ----------
    adata : AnnData object

    pos : list, optional
        Pass a list of markers that should be expressed in the resultant cells. The default is None.
    neg : list, optional
        Pass a list of markers that should not be expressed in the resultant cells. The default is None.
    classify_label : string, optional
        Provide a name for the calssified cells. The default is 'passed_classify'.
    subclassify_phenotype : list, optional
        If only a subset of phenotypes require to classified, pass the name of those phenotypes as a list
        through this argument. The default is None.
    phenotype : string, required
        Column name of the column containing the phenotype information. 
        This is important if `subclassify_phenotype` or `collapse_failed` arguments are used.
        The default is 'phenotype'.
    collapse_failed : bool, optional
        If set to true, the cells that were not classified based on the given criteria will be
        binned into a single category named 'failed_classify'. When False, the phenotype
        inforamation for other cells will be borrowed from the `phenotype` argument. The default is True.
    label : string, optional
        Key for the returned data, stored in `adata.obs`. The default is "classify".


    Returns
    -------
    adata : AnnData
        Updated AnnData Object.
    
    
    Example
    -------
    # Classify all cells with both pos and neg markers (Identify cytotoxic T-cells)
    adata = sm.hl.classify(adata, pos=['CD3D','CD8A'], neg=['ASMA'])
    
    # Classify specific sub-types of cells
    adata = sm.hl.classify(adata, pos=['CD3D','FOXP3'], neg=['ASMA'], subclassify_phenotype=['T cells','Regulatory T cells'])
    
    # Classify specific sub-types of cells and borrow labels from another column
    adata = sm.hl.classify(adata, pos=['CD3D'], neg=['ASMA'], subclassify_phenotype=['T cells'], collapse_failed=False, phenotype='phenotype')

    """
    
    # clean the input
    if isinstance(pos, str):
        pos = [pos]
    if isinstance(neg, str):
        neg = [neg]
    if isinstance(subclassify_phenotype, str):
        subclassify_phenotype = [subclassify_phenotype]
    
    
    # Create a dataFrame with the necessary inforamtion
    data = pd.DataFrame(adata.X, index= adata.obs.index, columns = adata.var.index)
    meta = pd.DataFrame(adata.obs[phenotype])
    
    # if user requests to subset a specific phenotype   
    if subclassify_phenotype is not None:
        subset_index = meta[meta[phenotype].isin(subclassify_phenotype)].index
        data = data.loc[subset_index]
        
    # Subset cells that pass the pos criteria
    if pos is not None:
        for i in pos:
            # subset data
            data = data[data[i] >= 0.5]
    
    # Subset cells that pass the neg criteria 
    if neg is not None and not data.empty:
        for j in neg:
            # subset data
            data = data[data[j] < 0.5]
    
    # cells that passed the classify criteria
    if data.empty:
        raise TypeError("No cells were found to satisfy your `classify` criteria")
    else:
        classify_idx = data.index
        classified = pd.DataFrame(np.repeat(classify_label, len(classify_idx)), index = classify_idx, columns = [phenotype])
        
        
    if collapse_failed is True:
        meta = meta.merge(classified, how='outer', left_index=True, right_index=True)
        meta.columns=[label,phenotype]
        meta[phenotype] = meta[phenotype].fillna('failed_classify')
    else:
        meta.update(classified)
    
    # Add to Anndata
    meta = meta.reindex(adata.obs.index)
    adata.obs[label] = meta[phenotype]
    
    # return
    return adata




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scimap',
 'scimap.archive',
 'scimap.helpers',
 'scimap.plotting',
 'scimap.preprocessing',
 'scimap.tests',
 'scimap.tools']

package_data = \
{'': ['*'], 'scimap.tests': ['_data/*']}

install_requires = \
['PhenoGraph>=1.5.7,<2.0.0',
 'TiffFile>=2020.11.18,<2021.0.0',
 'anndata>=0.7.4,<0.8.0',
 'dask>=2.30.0,<3.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'mkdocs>=1.1.2,<2.0.0',
 'napari>=0.4.0,<0.5.0',
 'numpy>=1.18.5,<2.0.0',
 'pandas>=1.0.4,<2.0.0',
 'plotly>=4.12.0,<5.0.0',
 'pytest-xvfb>=2.0.0,<3.0.0',
 'pytest>=5.4.3,<6.0.0',
 'scanpy>=1.6.0,<2.0.0',
 'scipy>=1.4.1,<2.0.0',
 'seaborn>=0.11.0,<0.12.0',
 'sklearn>=0.0,<0.1',
 'tifffile>=2020.6.3,<2021.0.0',
 'zarr>=2.5.0,<3.0.0']

entry_points = \
{'console_scripts': ['scimap = scimap.cli:main']}

setup_kwargs = {
    'name': 'scimap',
    'version': '0.9.7',
    'description': 'Single-Cell Image Analysis Package',
    'long_description': '# Single-Cell Image Analysis Package\n\nScimap is a scalable toolkit for analyzing single-cell multiplex imaging data. The package uses the [anndata](https://anndata.readthedocs.io/en/stable/anndata.AnnData.html) framework making it easy to integrate with other popular single-cell analysis toolkits such as [scanpy](https://scanpy.readthedocs.io/en/latest/#). It includes preprocessing, phenotyping, visualization, clustering, spatial analysis and differential spatial testing. The Python-based implementation efficiently deals with datasets of more than one million cells.\n\n\n[![Unix Build Status](https://img.shields.io/travis/ajitjohnson/scimap/master.svg?label=unix)](https://travis-ci.org/ajitjohnson/scimap)\n[![Windows Build Status](https://img.shields.io/appveyor/ci/ajitjohnson/scimap/master.svg?label=windows)](https://ci.appveyor.com/project/ajitjohnson/scimap)\n[![Documentation Status](https://readthedocs.org/projects/scimap-doc/badge/?version=latest)](https://scimap-doc.readthedocs.io/en/latest/?badge=latest)\n[![Downloads](https://pepy.tech/badge/scimap)](https://pepy.tech/project/scimap)\n[![PyPI Version](https://img.shields.io/pypi/v/scimap.svg)](https://pypi.org/project/scimap)\n[![PyPI License](https://img.shields.io/pypi/l/scimap.svg)](https://pypi.org/project/scimap)\n<!--[![Scrutinizer Code Quality](https://img.shields.io/scrutinizer/g/ajitjohnson/scimap.svg)](https://scrutinizer-ci.com/g/ajitjohnson/scimap/?branch=master)-->\n<!--[![Coverage Status](https://img.shields.io/coveralls/ajitjohnson/scimap/master.svg)](https://coveralls.io/r/ajitjohnson/scimap) -->\n\n## Installation\n\nWe strongly recommend installing `scimap` in a fresh virtual environment.\n\n```\n# If you have conda installed\nconda create --name scimap python=3.7\nconda activate scimap\n```\n\nInstall `scimap` directly into an activated virtual environment:\n\n```python\n$ pip install scimap\n$ pip install napari[all] # For visualizing images\n```\n\nAfter installation, the package can be imported as:\n\n```python\n$ python\n>>> import scimap as sm\n```\n\n## Get Started\n\n\n#### Detailed documentation of `scimap` functions and tutorials are available [here](https://scimap-doc.readthedocs.io/en/latest/).\n',
    'author': 'Ajit Johnson Nirmal',
    'author_email': 'ajitjohnson.n@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/scimap',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 16 08:34:04 2020
@author: Ajit Johnson Nirmal
Omero ROI to Scimap
"""

# Library
import pandas as pd
import numpy as np
import re
import matplotlib.path as mpath
from joblib import Parallel, delayed


def add_roi (adata, roi, x_coordinate='X_centroid',y_coordinate='Y_centroid',label='ROI'):
    
    """
    
    Parameters
    ----------
    adata : AnnData object

    roi : DataFrame
        Pandas dataframe of ROI's that have been extracted from Omero using the following script: https://gist.github.com/Yu-AnChen/58754f960ccd540e307ed991bc6901b0.
        Please note that the function currently does not handle overlapping ROI's and so make sure the ROI's are mutually exclusive.
    x_coordinate : float, required
        Column name containing the x-coordinates values. The default is 'X_centroid'.
    y_coordinate : float, required
        Column name containing the y-coordinates values. The default is 'Y_centroid'.
    label : string, optional
        Key for the returned data, stored in `adata.obs`. The default is 'ROI'.

    Returns
    -------
    adata
        Modified AnnData object. Check `adata.obs` for an additional column.
    
    Example
    -------
    roi = pd.read_csv('ROI/ROI_Z147_1_750.csv')
    adata = sm.hl.add_roi (adata, roi, label='aj_ROI')

    """
    
    # create data matrix that has the co-ordinates
    data = pd.DataFrame(adata.obs)[[x_coordinate, y_coordinate]]
    
    def add_roi_internal (roi_id):
        roi_subset = roi[roi['Id'] == roi_id]
        
        # Make a list of all ROI's from the ROI table
        all_points = [
        np.array(re.findall(r'\d+\.\d+', s))
            .astype(np.float64)
            .reshape(-1, 2)
        for s in roi_subset['all_points']
        ]
        
        # Identify the cells within the ROI
        path = mpath.Path(all_points[0], closed=True)
        
        # the mask to query cells within one ROI
        inside = data[path.contains_points(data)]
        inside['ROI'] = roi_subset['Name'].iloc[0]

        # return
        return inside
    
    # Apply function to all rows of the ROI dataframe
    roi_list = roi['Id'].unique()
    final_roi = Parallel(n_jobs=-1)(delayed(add_roi_internal)(roi_id=i) for i in roi_list)   
    
    # Merge all into a single DF
    final_roi = pd.concat(final_roi)[['ROI']]
    
    # Add the list to obs
    result = pd.merge(data, final_roi, left_index=True, right_index=True, how='outer')
    
    # Reindex
    result = result.reindex(adata.obs.index)
    result['ROI'] = result['ROI'].fillna('Other')
    
    # Add to adata
    adata.obs[label] = result['ROI']
    
    # return
    return adata
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 21 11:00:57 2020
@author: Ajit Johnson Nirmal
Heatmap plot to visualize spatial interaction output. The intensity represents 
number of interactions (scaled) observed and blank regions represent non-significant results.
"""

# Library
import seaborn as sns; sns.set(color_codes=True)
import numpy as np
import pandas as pd
import matplotlib
sns.set_style("white")

# Function
def spatial_interaction (adata, spatial_interaction='spatial_interaction',
                         summarize_plot=True, p_val=0.05,
                         row_cluster=False, col_cluster=False,
                         cmap = 'vlag', nonsig_color='grey', 
                         subset_phenotype=None, subset_neighbour_phenotype=None,
                         binary_view=False, **kwargs):
    """

    Parameters
    ----------
    adata : AnnData object

    spatial_interaction : string, optional
        In order to locate the spatial_interaction data within the AnnData object please provide the output 
        label/columnname of `sm.tl.spatial_interaction` function. The default is 'spatial_interaction'.
    summarize_plot : bool, optional
        In the event of analyzing multiple images, this argument allows users to
        plot the average cell-cell interaction across all images. The default is True.
    p_val : float, optional
        P-value cut-off above which interactions are not considered significant. The default is 0.05.
    row_cluster : bool, optional
        Cluster Rows. The default is False.
    col_cluster : bool, optional
        Cluster Columns. The default is False.
    subset_phenotype : list, optional
        If user requires to visualize a subset of phenotypes, it can be passed here. 
        e.g.  `subset_phenotype = ['celltype_A', 'celltype_B']`. The default is None.
    subset_neighbour_phenotype : list, optional
        If user requires to visualize a subset of interacting phenotypes, it can be passed here. 
        e.g.  `subset_neighbour_phenotype = ['celltype_C', 'celltype_D']`. The default is None.
    cmap : string, optional
        Color map to use for continous variables. 
        Can be a name or a Colormap instance (e.g. 'magma', 'viridis'). The default is 'vlag'.
    nonsig_color : string, optional
        Color for non-significant interactions (Interactions above the P-value cut-off will use this color).
        The default is 'grey'.
    binary_view : bool, optional
        Removes the intensity of intreaction and plots significant interactions and avoidance in a binary format.
        The default is 'False'.

    **kwargs : key:value pairs
        Are passed to sns.clustermap. Pass other parameters that works with `sns.clustermap`. e.g. `linecolor='black'`

    Example
    -------
    # spatial_interaction heatmap for a single image
    sm.pl.spatial_interaction(adata, summarize_plot=True, row_cluster=True, linewidths=0.75, linecolor='black')
    
    # spatial_interaction heatmap for multiple images
    sns.set(font_scale=0.6)
    sm.pl.spatial_interaction(adata, summarize_plot=False, row_cluster=True, col_cluster=True, yticklabels=True)

    """
    
    # set color for heatmap
    #cmap_updated = copy.copy(matplotlib.cm.get_cmap(cmap))
    cmap_updated = matplotlib.cm.get_cmap(cmap)
    cmap_updated.set_bad(color=nonsig_color)
    

    # Copy the interaction results from anndata object
    try:
        interaction_map = adata.uns[spatial_interaction].copy()
    except KeyError:
        raise ValueError('spatial_interaction not found- Please run sm.tl.spatial_interaction first')
        
    # subset the data if user requests
    if subset_phenotype is not None:
        if isinstance(subset_phenotype, str):
            subset_phenotype = [subset_phenotype]
        # subset the phenotype
        interaction_map = interaction_map[interaction_map['phenotype'].isin(subset_phenotype)]
    
    if subset_neighbour_phenotype is not None:
        if isinstance(subset_neighbour_phenotype, str):
            subset_neighbour_phenotype = [subset_neighbour_phenotype]
        # subset the phenotype
        interaction_map = interaction_map[interaction_map['neighbour_phenotype'].isin(subset_neighbour_phenotype)]
        
    # Seperate Interaction intensity from P-value
    p_value = interaction_map.filter(regex='pvalue_')    
    p_val_df = pd.concat([interaction_map[['phenotype','neighbour_phenotype']], p_value], axis=1, join='outer')
    p_val_df = p_val_df.set_index(['phenotype','neighbour_phenotype'])
    interaction_map = interaction_map[interaction_map.columns.difference(p_value.columns)]
    interaction_map = interaction_map.set_index(['phenotype','neighbour_phenotype'])
    
    # Binarize the values if user requests
    if binary_view == True:
        interaction_map[interaction_map > 0] = 1
        interaction_map[interaction_map <= 0] = -1
        

    if summarize_plot == True:
        # convert first two columns to multi-index column
        #interaction_map = interaction_map.set_index(['phenotype','neighbour_phenotype'])
        #p_val_df = p_val_df.set_index(['phenotype','neighbour_phenotype'])
        
        # If multiple images are present, take the average of interactions
        interaction_map['mean'] = interaction_map.mean(axis=1).values
        interaction_map = interaction_map[['mean']] # keep only the mean column
        interaction_map = interaction_map['mean'].unstack()
        # Do the same for P-values
        p_val_df['mean'] = p_val_df.mean(axis=1).values
        p_val_df = p_val_df[['mean']] # keep only the mean column
        # set the P-value threshold
        p_val_df.loc[p_val_df[p_val_df['mean'] > p_val].index,'mean'] = np.NaN       
        p_val_df = p_val_df['mean'].unstack()
        
        # Plotting heatmap
        mask = p_val_df.isnull() # identify the NAN's for masking 
        interaction_map = interaction_map.fillna(0) # replace nan's with 0 so that clustering will work
        # heatmap
        sns.clustermap(interaction_map, cmap=cmap, row_cluster=row_cluster, col_cluster=col_cluster,  mask=mask, **kwargs)
        
    else:
        if len(interaction_map.columns) <= 3:
            raise ValueError('Data for only a single image is available please set summarize_plot=True and try again')
        # convert first two columns to multi-index column
        #interaction_map = interaction_map.set_index(['phenotype','neighbour_phenotype'])
        #p_val_df = p_val_df.set_index(['phenotype','neighbour_phenotype'])
        
        # P value threshold
        p_val_df = p_val_df.apply(lambda x: np.where(x > p_val,np.nan,x))
        
        # Remove rows that are all nan
        idx = p_val_df.index[p_val_df.isnull().all(1)] # Find all nan rows
        interaction_map = interaction_map.loc[interaction_map.index.difference(idx)] # clean intensity data
        p_val_df = p_val_df.loc[p_val_df.index.difference(idx)] # clean p-value data
        
        # Plotting heatmap
        mask = p_val_df.isnull() # identify the NAN's for masking 
        interaction_map = interaction_map.fillna(0) # replace nan's with 0 so that clustering will work
        mask.columns = interaction_map.columns
        # Plot
        sns.clustermap(interaction_map, cmap=cmap, row_cluster=row_cluster, col_cluster=col_cluster, mask=mask, **kwargs)
        

